import os.path
import sys
import copy
from PyQt5.QtWidgets            import  QWidget, QTabWidget
from PyQt5.QtWidgets            import  QVBoxLayout
from PyQt5.QtWidgets            import  QApplication
from PyQt5.QtWidgets            import  QFileDialog
from PyQt5.QtCore               import  Qt
from PDielec.GUI.MainTab         import  MainTab
from PDielec.GUI.SettingsTab     import  SettingsTab
from PDielec.GUI.ScenarioTab     import  ScenarioTab
from PDielec.GUI.PlottingTab     import  PlottingTab
from PDielec.GUI.AnalysisTab     import  AnalysisTab
from PDielec.GUI.ViewerTab       import  ViewerTab
from PDielec.GUI.FitterTab       import  FitterTab
from PDielec.Utilities           import  Debug

class NoteBook(QWidget):

    def __init__(self, parent, program, filename, spreadsheet, debug=False, progressbar=False, scripting=False, ncpus=0, threading=False):
        super(QWidget, self).__init__(parent)
        global debugger
        debugger = Debug(debug,'NoteBook:')
        self.reader = None
        self.progressbar=progressbar
        self.spreadsheet = None
        self.threading = threading
        self.ncpus = ncpus
        self.scripting = scripting
        self.debug = debug
        self.plottingCalculationRequired = True
        self.analysisCalculationRequired = True
        self.visualerCalculationRequired = True
        self.fittingCalculationRequired = True
        self.old_tab_index = None
        self.layout = QVBoxLayout()
        # The number of tabs before we have scenarios
        self.tabOffSet = 2
        #
        # Initialize tab screen
        #
        self.tabs = QTabWidget(self)
        self.tabs.currentChanged.connect(self.on_tabs_currentChanged)
        self.mainTab = MainTab(self, program, filename, spreadsheet, debug=debug)
        self.settingsTab = SettingsTab(self, debug=debug)
        if filename != '' and not self.scripting:
            debugger.print('Refreshing settingsTab in notebook initialisation - filename',filename)
            self.settingsTab.refresh()
        #
        # Open more windows
        #
        self.scenarios = []
        self.scenarios.append( ScenarioTab(self, debug=debug ) )
        self.scenarios[0].setScenarioIndex(0)
        #
        # Open the plotting tab
        #
        self.plottingTab = PlottingTab(self, debug=debug)
        if filename != '' and not self.scripting:
            debugger.print('Refreshing plotting because filename is set')
            self.plottingTab.refresh()
        #
        # Open the Analysis tab
        #
        self.analysisTab = AnalysisTab(self, debug=debug)
        if filename != '' and not self.scripting:
            debugger.print('Refreshing analysis because filename is set')
            self.analysisTab.refresh()
        #
        # Open the Viewer tab
        #
        self.viewerTab = ViewerTab(self, debug=debug)
        #
        # Open the Fitter tab
        #
        self.fitterTab = FitterTab(self, debug=debug)
        #
        # Add tabs
        #
        self.tabs.addTab(self.mainTab,'Main')
        self.tabs.addTab(self.settingsTab,'Settings')
        for i,tab in enumerate(self.scenarios):
            self.tabs.addTab(tab,'Scenario '+str(i+1))
        self.tabs.addTab(self.plottingTab,'Plotting')
        self.tabs.addTab(self.analysisTab,'Analysis')
        self.tabs.addTab(self.viewerTab,'3D Viewer')
        self.tabs.addTab(self.fitterTab,'Fitter')

        # Add the tab widget
        self.layout.addWidget(self.tabs)
        self.setLayout(self.layout)
        return

    def addScenario(self,copyFromIndex=-2):
        debugger.print('Settings for scenario', copyFromIndex)
        self.plottingCalculationRequired = True
        self.scenarios.append( ScenarioTab(self, self.debug) )
        self.scenarios[-1].settings = copy.deepcopy(self.scenarios[copyFromIndex].settings)
        # debugger.print('Settings for new scenario')
        self.scenarios[-1].refresh(force=True)
        for i,scenario in enumerate(self.scenarios):
            scenario.setScenarioIndex(i)
        n = len(self.scenarios)
        self.tabs.insertTab(self.tabOffSet+n-1,self.scenarios[-1],'Scenario '+str(n))
        self.tabs.setCurrentIndex(self.tabOffSet+n-1)
        return

    def print_settings(self):
        # Print the settings of all the settings that have been used to a file settings.py
        qf = QFileDialog()
        qf.setWindowTitle('Save the program settings to a file')
        filename,selection = qf.getSaveFileName()
        if filename == '':
            return
        print('Current settings will be saved to '+filename)
        fd = open(filename,'w')
        ntabs = 2+ len(self.scenarios) + 4
        self.print_tab_settings(self.mainTab, 'mainTab',fd)
        print('tab.refresh(force=True)',file=fd)
        self.print_tab_settings(self.settingsTab, 'settingsTab',fd)
        print('tab.sigmas_cm1 =',self.settingsTab.sigmas_cm1,file=fd)
        print('tab.refresh(force=True)',file=fd)
        requireNewScenario = False
        for i,tab in enumerate(self.scenarios):
            self.print_tab_settings(tab, 'scenarios[{}]'.format(i), fd, new_scenario = requireNewScenario)
            print('tab.refresh(force=True)',file=fd)
            requireNewScenario = True
        self.print_tab_settings(self.plottingTab, 'plottingTab',fd)
        print('tab.refresh(force=True)',file=fd)
        self.print_tab_settings(self.analysisTab, 'analysisTab',fd)
        print('tab.refresh(force=True)',file=fd)
        self.print_tab_settings(self.viewerTab, 'viewerTab',fd)
        print('tab.refresh(force=True)',file=fd)
        self.print_tab_settings(self.fitterTab, 'fitterTab',fd)
        print('tab.refresh(force=True)',file=fd)
        fd.close()
        return

    def print_tab_settings(self,tab,title,fd,new_scenario = False):
        print('#',file=fd)
        print('#',file=fd)
        if new_scenario:
            print('self.notebook.addScenario()',file=fd)
        print('tab = self.notebook.'+title,file=fd)
        for item in tab.settings:
            value = tab.settings[item]
            if 'str' in str(type(value)):
                print('tab.settings[\''+item+'\'] = \'{}\''.format(tab.settings[item]),file=fd)
            else:
                print('tab.settings[\''+item+'\'] = ', tab.settings[item],file=fd)

    def deleteScenario(self,index):
        # Don't delete the last scenario
        if len(self.scenarios) > 1:
            self.plottingCalculationRequired = True
            self.tabs.removeTab(self.tabOffSet+index)
            del self.scenarios[index]
            for i,scenario in enumerate(self.scenarios):
                scenario.setScenarioIndex(i)
                self.tabs.setTabText(self.tabOffSet+i,'Scenario '+str(i+1))
        return

    def refresh(self,force=False):
        if self.scripting:
            debugger.print('Notebook aborting refresh because of scripting')
            return
        debugger.print('Notebook refresh changed',force)
        ntabs = 2 + len(self.scenarios) + 4
        self.mainTab.refresh(force=force)
        self.settingsTab.refresh(force=force)
        for tab in self.scenarios:
            tab.refresh(force=force)
        self.tabs.setCurrentIndex(ntabs-5)
        self.plottingTab.refresh(force=force)
        self.tabs.setCurrentIndex(ntabs-4)
        self.analysisTab.refresh(force=force)
        self.tabs.setCurrentIndex(ntabs-3)
        self.viewerTab.refresh(force=force)
        self.tabs.setCurrentIndex(ntabs-2)
        self.fitterTab.refresh(force=force)
        self.tabs.setCurrentIndex(ntabs-1)

    def write_spreadsheet(self):
        debugger.print('Write spreadsheet')
        self.mainTab.write_spreadsheet()
        self.settingsTab.write_spreadsheet()
        self.analysisTab.write_spreadsheet()
        self.plottingTab.write_spreadsheet()

    def on_tabs_currentChanged(self, tabindex):
        debugger.print('Tab index changed', tabindex)
        if self.scripting:
            return
        # See if we have to up date a tab we have left
        if self.old_tab_index is not None:
            if self.old_tab_index == 0:
                self.mainTab.refresh()
            elif self.old_tab_index == 1:
                self.settingsTab.refresh()
        # end if
        #       Number of tabs
        ntabs = 2 + len(self.scenarios) + 4
        if tabindex == ntabs-1:
            # fitter tab
            self.fitterTab.refresh()
        elif tabindex == ntabs-2:
            # viewer tab
            self.viewerTab.refresh()
        elif tabindex == ntabs-3:
            # analysis tab
            self.analysisTab.refresh()
        elif tabindex == ntabs-4:
            # plottings tab
            self.plottingTab.refresh()
        self.old_tab_index = tabindex

    def keyPressEvent(self, e):
        if (e.key() == Qt.Key_S)  and QApplication.keyboardModifiers() and Qt.ControlModifier:
            print('Control S has been pressed')
            self.print_settings()
        elif (e.key() == Qt.Key_C)  and QApplication.keyboardModifiers() and Qt.ControlModifier:
            print('Control C has been pressed')
            print('The program will close down')
            sys.exit()



import xlsxwriter as xlsx
class SpreadSheetManager():
    def __init__(self, filename):
        # Manage a spread sheet for PDielec / PDGui
        self.workbook = xlsx.Workbook(filename)
        self.tab_names = ['Main', 'Settings', 'Scenarios', 'Molar Absorption','Absorption', 'Real Permittivity', 'Imaginary Permittivity', 'ATR Reflectance','Analysis']
        self.worksheets = {}
        # Positions points to where we write to next
        self.positions  = {}
        self.max_col    = {}
        self.max_row    = {}
        for tab in self.tab_names:
            self.worksheets[tab] = self.workbook.add_worksheet(tab)
            self.positions[tab] = (0,0)
            self.max_col[tab] = 0
            self.max_row[tab] = 0
        self.name = 'Main'

    def selectWorkSheet(self,name):
        self.name = name

    def writeNextRow(self,items, row=None, col=None, check=''):
        oldRow,oldCol = self.positions[self.name]
        if col is None:
            col = oldCol
        if row is None:
            row = oldRow
        if col == 0:
            print('We have a problem, col is 0')
        self.write(row,0,check)
        for item in items:
            #print('writing ', row, col, item)
            self.write(row, col, item)
            col += 1
        row += 1

    def write(self,row,col,item):
        self.worksheets[self.name].write(row,col,item)
        self.max_col[self.name] = max(self.max_col[self.name], col)
        self.max_row[self.name] = max(self.max_row[self.name], row)
        self.positions[self.name] = (row+1,col+1)

    def delete(self):
        for row in range(0,self.max_row[self.name]):
            for col in range(0, self.max_col[self.name]):
                self.worksheets[self.name].write(row,col,'')
        self.positions[self.name] = (-1,0)
        self.max_col[self.name] = 0
        self.max_row[self.name] = 0

    def close(self):
        self.workbook.close()



#!/usr/bin/python
#
# Copyright 2015 John Kendrick
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License
# along with this program, if not see https://opensource.org/licenses/MIT
#
"""Read the contents of a directory containg VASP input and output files"""
import re
import os
import numpy as np
from PDielec.UnitCell import UnitCell
from PDielec.GenericOutputReader import GenericOutputReader
from PDielec.Constants           import atomic_number_to_element


def myfloat(string):
    '''A replacement for float() which will return a large number if it finds a * in the string '''
    if '*' in string:
        return 9999.999
    else:
        return float(string)


class VaspOutputReader(GenericOutputReader):
    """Read the contents of a directory containg VASP input and output files"""

    def __init__(self, names):
        GenericOutputReader.__init__(self, names)
        self.type                    = 'Vasp output'
        self._pspots                  = {}
        self._ediff                   = 0.0
        self._pulay                  = None
        self._ibrion                  = 0
        self._potim                   = 0.0
        return

    def _read_output_files(self):
        """Read the vasp files in the directory"""
        self.manage = {}   # Empty the dictionary matching phrases
        self.manage['ionspertype']  = (re.compile('   ions per type ='), self._read_ionspertype)
        # self.manage['masses_skip']  = (re.compile('  Mass of Ions in am'), self._read_skip4)
        self.manage['pspots']       = (re.compile(' POTCAR:'), self._read_pspot)
        self.manage['arrays']       = (re.compile(' Dimension of arrays:'), self._read_array_dimensions)
        #self.manage['masses']       = (re.compile('   POMASS ='), self._read_masses)
        self.manage['spin']         = (re.compile('   ISPIN  = '), self._read_spin)
        self.manage['encut']        = (re.compile('   ENCUT  = '), self._read_encut)
        self.manage['ediff']        = (re.compile('   EDIFF  = '), self._read_ediff)
        self.manage['ibrion']       = (re.compile('   IBRION = '), self._read_ibrion)
        self.manage['potim']        = (re.compile('   POTIM  = '), self._read_potim)
        self.manage['nelect']       = (re.compile('   NELECT = '), self._read_nelect)
        self.manage['lattice']      = (re.compile('  volume of cell :'), self._read_lattice_vectors)
        self.manage['fractional']   = (re.compile(' position of ions in fractional coordinates'), self._read_fractional_coordinates)
        self.manage['forces']       = (re.compile(' POSITION  *TOTAL-FORCE'), self._read_forces)
        self.manage['energy']       = (re.compile('  FREE ENERGIE OF THE ION'), self._read_energy)
        self.manage['magnet']       = (re.compile(' number of electron '), self._read_magnet)
        self.manage['pressure']     = (re.compile('  external pressure ='), self._read_external_pressure)
        self.manage['skip1']        = (re.compile(' old parameters found'), self._read_skip4)
        self.manage['staticDielectric']  = (re.compile(' MACROSCOPIC STATIC DIELECTRIC TENSOR .including'), self._read_static_dielectric)
        self.manage['staticIonic']  = (re.compile(' MACROSCOPIC STATIC DIELECTRIC TENSOR IONIC CONT'), self._read_ionic_dielectric)
        self.manage['bornCharges']  = (re.compile(' BORN EFFECTIVE CHARGES'), self._read_born_charges)
        self.manage['eigenvectors'] = (re.compile(' Eigenvectors and eige'), self._read_eigenvectors)
        self.manage['eigenskip']    = (re.compile(' Eigenvectors after division'), self._read_skip4)
        self.manage['elastic']      = (re.compile(' TOTAL ELASTIC MODULI'), self._read_elastic_constants)
        self.manage['kpoint']      = (re.compile('^Gamma'), self._read_kpoint_grid)
        self.manage['species']      = (re.compile('^ *Atomic configuration'), self._read_species)
        self.manage['newmasses']  = (re.compile('  Mass of Ions in am'), self._read_newmasses)
        for f in self._outputfiles:
            self._read_output_file(f)
        return

    def _read_forces(self, line):
        line = self.file_descriptor.readline()
        maxf = 0.0
        rmsf = 0.0
        for i in range(self.nions):
            line = self.file_descriptor.readline()
            forces = [ float(f) for f in line.split()[3:6] ]
            for f in forces:
                rmsf += f*f
                if abs(f) > maxf:
                    maxf = abs(f)
                # end if
            # end for f
        #end for i
        if not "max_force" in self.iterations:
            self.iterations["max_force"] = []
        self.iterations["max_force"].append(maxf)
        if not "rms_force" in self.iterations:
            self.iterations["rms_force"] = []
        self.iterations["rms_force"].append(rmsf)
        return

    def _read_species(self, line):
        line = self.file_descriptor.readline()
        nlines = int(line.split()[0])
        line = self.file_descriptor.readline()
        zcharge = 0.0
        for i in range(nlines):
            line = self.file_descriptor.readline()
            zcharge = zcharge + float(line.split()[4])
        self.species.append(atomic_number_to_element[int(zcharge+0.001)])
        self.nspecies = len(self.species)
        return

    def _read_kpoint_grid(self, line):
        line = self.file_descriptor.readline()
        self.kpoint_grid = [int(f) for f in line.split()[0:3] ]
        return

    def _read_ionspertype(self, line):
        self.ions_per_type = [int(i) for i in line.split()[4:]]
        self.nspecies = len(self.ions_per_type)
        return

    def _read_newmasses(self, line):
        self.masses_per_type = []
        line = self.file_descriptor.readline()
        mass_string = line[12:]
        start = 0
        increment = 6
        for i in range(self.nspecies):
            mass = mass_string[start:start+increment]
            self.masses_per_type.append(float(mass))
            start = start + increment
        self.masses = []
        self.atom_type_list = []
        self.species_list = []
        for k, mass in enumerate(self.masses_per_type):
            n = self.ions_per_type[k]
            for i in range(n):
                self.atom_type_list.append(k)
                self.masses.append(mass)
                self.species_list.append(self.species[k])
            # end loop over i
        # end look over current know types
        return

    def _read_masses(self, line):
        mass_string = line.split()[2]
        mass_string = mass_string.replace(";", "")
        self.masses_per_type.append(float(mass_string))
        return

    def _read_eigenvectors(self, line):
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.frequencies = []
        self.mass_weighted_normal_modes = []
        n = 3 * self.nions
        for i in range(n):
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            imaginary = (line.split()[1] == "f/i=")
            if imaginary:
                # represent imaginary by negative real
                freq = -float(line.split()[6])
            else:
                freq = float(line.split()[7])
            # end if
            self.frequencies.append(freq)
            line = self.file_descriptor.readline()
            a = []
            for j in range(self.nions):
                line = self.file_descriptor.readline()
                a.append([float(line.split()[3]), float(line.split()[4]), float(line.split()[5])])
            # end for j
            self.mass_weighted_normal_modes.append(a)
        # end of for i in range(n)
        return

    def _read_born_charges(self, line):
        """Read the born charges from the OUTCAR file.
           Each row of the output refers to a given field direction
           Each column in the row refers the atomic displacement
           so the output is arranged [[a1x a1y a1z]
                                      [ a2x a2y a2z]
                                      [ a3x a3y a3z]]
           where 1,2,3 are the field directions and x, y, z are the atomic displacements"""
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.born_charges = []
        for i in range(self.nions):
            line = self.file_descriptor.readline()
            b = []
            b.append([float(line.split()[1]), float(line.split()[2]), float(line.split()[3])])
            line = self.file_descriptor.readline()
            b.append([float(line.split()[1]), float(line.split()[2]), float(line.split()[3])])
            line = self.file_descriptor.readline()
            b.append([float(line.split()[1]), float(line.split()[2]), float(line.split()[3])])
            line = self.file_descriptor.readline()
            self.born_charges.append(b)
        return

    def _read_elastic_constants(self, line):
        # Read the total elastic constants
        elastic_constants = []
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        line = self.file_descriptor.readline()
        elastic_constants.append([myfloat(f) for f in line.split()[1:7]])
        econs = np.array(elastic_constants)
        # convert from kBar to GPa
        econs = econs / 10.0
        self.elastic_constants = econs.tolist()
        return

    def _read_ionic_dielectric(self, line):
        # Read the ionic contribution to the static dielectric and use it to computet
        # the full static dielectric constant
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        # the is zero frequency ionic contribution to the static dielectric
        ionic_dielectric = []
        if '*' in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        ionic_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if '*' in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        ionic_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if '*' in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        ionic_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        array1 = np.array(self.zerof_optical_dielectric)
        array2 = np.array(ionic_dielectric)
        array3 = array1 + array2
        self.zerof_static_dielectric = array3.tolist()
        return

    def _read_static_dielectric(self, line):
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        # the is epsilon infinity
        self.zerof_optical_dielectric = []
        if '*' in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        self.zerof_optical_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if '*' in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        self.zerof_optical_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        line = self.file_descriptor.readline()
        if '*' in line or len(line.split()) < 3:
            line = "99999.999 99999.999 99999.999"
        self.zerof_optical_dielectric.append([myfloat(f) for f in line.split()[0:3]])
        return

    def _read_skip4(self, line):
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        return

    def _read_external_pressure(self, line):
        # Vasp writes out kbar so convert to GPa
        self.pressure = float(line.split()[3])/10.0
        self._pulay = float(line.split()[8])/10.0
        return

    def _read_pspot(self, line):
        self._pspots[line.split()[2]] = line.split()[1]
        return

    def _read_array_dimensions(self, line):
        line = self.file_descriptor.readline()
        self.kpoints = int(line.split()[3])
        self.nbands = int(line.split()[14])
        line = self.file_descriptor.readline()
        self.nions = int(line.split()[11])
        return

    def _read_lattice_vectors(self, line):
        self.volume = float(line.split()[4])
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        avector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        bvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        cvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        cell = UnitCell(avector, bvector, cvector)
        if self.ncells > 0:
            cell.set_fractional_coordinates(self.unit_cells[-1].fractional_coordinates)
            cell.set_element_names(self.species_list)
        self.unit_cells.append(cell)
        self.ncells = len(self.unit_cells)
        return

    def _read_fractional_coordinates(self, line):
        n = 0
        ions = []
        for n in range(self.nions):
            line = self.file_descriptor.readline()
            ions.append([float(f) for f in line.split()[0:3]])
        self.unit_cells[-1].set_fractional_coordinates(ions)
        self.unit_cells[-1].set_element_names(self.species_list)
        return

    def _read_spin(self, line):
        self.spin = int(line.split()[2])
        return

    def _read_encut(self, line):
        self.energy_cutoff = float(line.split()[2])
        return

    def _read_ediff(self, line):
        self._ediff = float(line.split()[2])
        return

    def _read_ibrion(self, line):
        self._ibrion = int(line.split()[2])
        return

    def _read_potim(self, line):
        self._potim = float(line.split()[2])
        return

    def _read_nelect(self, line):
        self.electrons = int(float(line.split()[2]))
        return

    def _read_magnet(self, line):
        self.electrons = float(line.split()[3])
        if len(line.split()) > 5:
            self.magnetization = float(line.split()[5])
        else:
            self.magnetization = 0.0

    def _read_energy(self, line):
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.final_free_energy = float(line.split()[4])
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        self.final_energy_without_entropy = float(line.split()[3])
        return



#!/usr/bin/python
#
# Copyright 2015 John Kendrick
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License
# along with this program, if not see https://opensource.org/licenses/MIT
#
"""Generic reader of output files.  Actual reader should inherit from this class"""
from __future__ import print_function
import math
import os
import sys
import string
import numpy as np
from PDielec.Constants  import wavenumber, avogadro_si, atomic_number_to_element, amu
from PDielec.Plotter    import print3x3, print_reals, print_strings, print_ints
from PDielec.Calculator import cleanup_symbol

class GenericOutputReader:
    """Generic reader of output files.  Actual reader should inherit from this class"""

    def __init__(self, filenames):
        self._outputfiles               = filenames
        self.names                      = [os.path.abspath(f) for f in filenames]
        self.debug                      = False
        self.type                       = 'Unkown'
        self.ncells                     = 0
        self.unit_cells                 = []
        self.nsteps                     = 0
        self.electrons                  = 0
        self.spin                       = 0
        self.nbands                     = 0
        self.volume                     = 0
        self.nions                      = 0
        self.nspecies                   = 0
        self.species                    = []
        self.final_free_energy          = 0.0
        self.final_energy_without_entropy = 0.0
        self.kpoints                    = 1
        self.kpoint_grid                = [ 1, 1, 1 ]
        self.energy_cutoff              = 0.0
        self.born_charges               = []
        self.manage                     = {}
        self.iterations                 = {}
        self.file_descriptor            = ''
        self.pressure                   = 0
        self.magnetization              = 0.0
        # this in epsilon infinity
        self.zerof_optical_dielectric   = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        # this is the zero frequency static dielectric constant
        self.zerof_static_dielectric    = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        self.elastic_constants          = [[0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0]]
        self.frequencies                = []
        self.mass_weighted_normal_modes = []
        self.ions_per_type              = []
        self.atom_type_list             = []
        self.masses                     = []
        self.masses_per_type            = []
        self.program_mass_dictionary    = {}
        self.eckart                     = False
        self.hessian_symmetrisation     = "symm"
        self.open_filename              = ""
        self.open_directory             = ""
        self._old_masses                = []
        self.nomass_hessian             = None
        self.nomass_hessian_has_been_set= False
        self.original_born_charges      = None
        self.original_born_charges_are_being_used = True
        return

    def read_output(self):
        """Interface to the private read output files methods"""
        self._read_output_files()
        return

    def reset_masses(self):
        #  If the mass needs reseting use the original (program) mass dictionary
        mass_dictionary = {}
        if self.debug:
            print("Re setting mass dictionary to the program values")
        if self.program_mass_dictionary:
            self.change_masses(self.program_mass_dictionary,mass_dictionary)

    def getSpecies(self):
        return [ cleanup_symbol(el) for el in self.species ]

    def mass_dictionary(self):
        dictionary = {}
        for symbol,mass in zip(self.species,self.masses_per_type):
            # the element name may be appended with a digit or an underscore
            element = cleanup_symbol(symbol)
            dictionary[element] = mass
        if self.debug:
            print("new mass_dictionary", dictionary)
        return dictionary

    def change_masses(self, new_masses, mass_dictionary):
        # Change the masses of the species stored, using the new_masses dictionary
        # if any of the elements are in the mass_dictionary then use that mass instead
        if not self.program_mass_dictionary:
            # We only want to do this once - remember the program masses as a dictionary
            if self.debug:
                print("Setting program mass dictionary")
            for symbol,mass in zip(self.species,self.masses_per_type):
                element = cleanup_symbol(symbol)
                self.program_mass_dictionary[element] = mass
        if self.debug:
            print("changing masses", self.program_mass_dictionary)
        self.masses = []
        self.masses_per_type = []
        for symbol in self.species:
            # the element name may be appended with a digit or an underscore
            element = cleanup_symbol(symbol)
            mass = new_masses[element]
            if element in mass_dictionary:
                mass = mass_dictionary[element]
            self.masses_per_type.append(mass)
        # end for symbol
        self.masses = [ self.masses_per_type[atype] for atype in self.atom_type_list ]
        if self.debug:
            print("new masses", self.masses)
        return

    def print_info(self):
        """Print information about the system"""
        # Generic printing of information
        print("")
        print("Summary of information contained in the QM/MM Reader")
        print("")
        print("Number of atoms: {:5d}".format(self.nions))
        print("")
        print("Number of species: {:5d}".format(self.nspecies))
        print_strings("Species:", self.species)
        print_ints("Number of atoms for each species:", self.ions_per_type)
        print_reals("Mass of each species:", self.masses_per_type,format="{:10.6f}")
        print_ints("Atom type list:", self.atom_type_list)
        print("")
        print("Number of kpoints: {:5d}".format(self.kpoints))
        print("")
        print("Kpoint grid      : {:5d} {:5d} {:5d}".format(self.kpoint_grid[0], self.kpoint_grid[1], self.kpoint_grid[2]))
        print("")
        print("Energy cutoff (eV): {:f}".format(self.energy_cutoff))
        print("")
        print_reals("Frequencies (cm-1):", self.frequencies)
        print_reals("Masses (amu):", self.masses,format="{:10.6f}")
        for i, charges in enumerate(self.born_charges):
            title = "Born Charges for Atom {:d}".format(i)
            print3x3(title, charges)
        print3x3("Epsilon inf: ", self.zerof_optical_dielectric)
        print3x3("Unit cell: ", self.unit_cells[-1].lattice)
        print(" ")
        print("Volume of cell: {:f}".format(self.volume))
        mtotal = 0.0
        for m in self.masses:
            mtotal = mtotal + m
        print("Total mass is: {:f} g/mol".format(mtotal))
        print("Density is: {:f} g/cc".format(mtotal/(avogadro_si * self.volume * 1.0e-24)))
        print(" ")
        return

    def _read_output_files(self):
        """Read the file names"""
        # Define the search keys to be looked for in the files
        print("Error _read_output_files must be defined by the actual file reader")
        return

    def _read_output_file(self, name):
        """Read through the files for key words.  The keywords are established in _read_output_files"""
        # Check to see if the file exists....
        if not os.path.isfile(name):
            print("Warning file is not present: ", name, file=sys.stderr)
            return
        # Open file and store file name and directory
        self.file_descriptor = open(name, 'r')
        self.open_filename = name
        self.open_directory = os.path.dirname(name)
        if self.open_directory == "":
            self.open_directory = "."
        # Loop through the contents of the file a line at a time and parse the contents
        line = self.file_descriptor.readline()
        while line != '':
            for k in self.manage:
                if self.manage[k][0].match(line):
                    method   = self.manage[k][1]
                    if self.debug:
                        print('_read_output_file: Match found {}'.format(k))
                    method(line)
                    break
                # end if
            # end for
            line = self.file_descriptor.readline()
        # end while
        self.file_descriptor.close()
        return

    def _symmetric_orthogonalisation(self, A):
        """Private routine to perform symetric orthogonalisation"""
        # The matrix A is only approximately orthogonal
        n = np.size(A, 0)
        unity = np.eye(n)
        Ak = A
        for k in range(3):
            Bk = np.dot(Ak, Ak.T)
            error = unity - Bk
            Ck = np.linalg.inv(unity + Bk)
            Kk = np.dot(error, Ck)
            Ak = np.dot((unity + Kk), Ak)
            error  = np.sum(np.abs(error))
            if self.debug:
                print("Orthogonalisation iteration: ", error)
        # end for k
        return Ak

    def calculate_mass_weighted_normal_modes(self):
        """Calculate the mass weighted normal modes"""
        #
        # Reconstruct the massweighted hessian
        # If necessary diagonalise it and extract the frequencies and normal modes
        # Any projection can be performed here
        #
        # First step is to reconstruct the dynamical matrix (D) from the frequencies and the eigenvectors
        # f^2 = UT . D . U
        # and U is a hermitian matrix so U-1 = UT
        # D = (UT)-1 f^2 U-1 = U f UT
        # Construct UT from the normal modes
        if self.debug:
            print("calculate mass weighted normal modes")
        n = np.size(self.mass_weighted_normal_modes, 0)
        m = np.size(self.mass_weighted_normal_modes, 1)*3
        nmodes = 3*self.nions
        UT = np.zeros((n, m))
        frequencies_a = np.array(self.frequencies) * wavenumber
        if self.debug:
            print("frequencies_a",frequencies_a)
        masses = np.array(self.masses)*amu
        # if the non mass-weighted hasn't been set, set it
        if not self.nomass_hessian_has_been_set:
            if self.debug:
                print("hessian was not set")
            self.nomass_hessian_has_been_set = True
            for imode, mode in enumerate(self.mass_weighted_normal_modes):
                n = 0
                for atom in mode:
                    # in python the first index is the row of the matrix, the second is the column
                    UT[imode, n+0] = atom[0]
                    UT[imode, n+1] = atom[1]
                    UT[imode, n+2] = atom[2]
                    n = n + 3
                # end for atom
            # end for imode
            # convert the frequencies^2 to a real diagonal array
            # Warning we have to make sure the sign is correct here
            # The convention is that if the frequency is negative
            # then it is really imaginary, so the square of the frequency
            # will be negative too.
            frequencies_a = np.array(self.frequencies) * wavenumber
            f2 = np.diag(np.sign(frequencies_a)*np.real(frequencies_a*frequencies_a))
            # The back transformation uses approximately orthogonal (unitary) matrices because of rounding issues on reading vectors
            # So before that lets orthogonalise them
            UT = self._symmetric_orthogonalisation(UT)
            hessian = np.dot(np.dot(UT.T, f2), UT)
            # Make sure the dynamical matrix is real
            hessian = np.real(hessian)
            # We are going to store the non mass-weighted hessian, since the frequencies came from the MM/QM program
            # they are calculated using the program masses
            current_mass_dictionary = self.mass_dictionary()
            if self.debug:
                print("current mass dictionary", current_mass_dictionary)
            # There is a chance that the program_mass_dictionary hasn't been set - if it hasn't use the current masses
            if not self.program_mass_dictionary:
                self.program_mass_dictionary = current_mass_dictionary
            if self.debug:
                print("program mass dictionary", self.program_mass_dictionary)
            self.change_masses(self.program_mass_dictionary, {})
            masses = np.array(self.masses)*amu
            # remove the mass weighting from the hessian and store
            self.nomass_hessian = self._remove_mass_weighting(hessian,masses)
            # finally replace the masses with those set before we did this
            self.change_masses(current_mass_dictionary, {})
            if self.debug:
                print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        # If the masses have been changed then alter the mass weighted hessian here
        masses = np.array(self.masses)*amu
        if self.debug:
            print("masses", masses)
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        hessian = self._modify_mass_weighting(self.nomass_hessian, masses)
        if self.debug:
            print("mass weighted hessian", hessian[0:4][0])
        # Project out the translational modes if requested
        if self.eckart:
            hessian = self.project(hessian)
            if self.debug:
                print("projected hessian", hessian[0:4][0])
        # Find its eigenvalues and eigen vectors
        eig_val, eig_vec = np.linalg.eigh(hessian)
        self.mass_weighted_normal_modes = []
        # Store the new frequencies, using the negative convention for imaginary modes
        for i in range(nmodes):
            if eig_val[i] < 0:
                frequencies_a[i] = -math.sqrt(-eig_val[i]) / wavenumber
            else:
                frequencies_a[i] = math.sqrt(eig_val[i]) / wavenumber
        self.frequencies = frequencies_a.tolist()
        if self.debug:
            print("calculated frequencies", self.frequencies)
        # Store the mass weighted normal modes
        for i in range(nmodes):
            mode = []
            n = 0
            for j in range(self.nions):
                modea = [eig_vec[n][i], eig_vec[n+1][i], eig_vec[n+2][i]]
                n = n + 3
                mode.append(modea)
            self.mass_weighted_normal_modes.append(mode)
        # end for i
        return self.mass_weighted_normal_modes

    def project(self, hessian):
        """Apply projection operators to remove translation
        Take the given matrix (np.array)
        Project out the translational modes"""
        #
        new_hessian = np.zeros_like(hessian)
        nmodes = self.nions*3
        unit = np.eye(nmodes)
        p1 = np.zeros(nmodes)
        p2 = np.zeros(nmodes)
        p3 = np.zeros(nmodes)
        for i in range(self.nions):
            p1[i*3+0] = math.sqrt(self.masses[i])
            p2[i*3+1] = math.sqrt(self.masses[i])
            p3[i*3+2] = math.sqrt(self.masses[i])
        # end for i
        # Normalise
        p1 = p1 / math.sqrt(np.dot(p1, p1))
        p2 = p2 / math.sqrt(np.dot(p2, p2))
        p3 = p3 / math.sqrt(np.dot(p3, p3))
        # Form the projection operators
        P1 = unit - np.outer(p1, p1)
        P2 = unit - np.outer(p2, p2)
        P3 = unit - np.outer(p3, p3)
        # Now project out
        new_hessian = np.dot(np.dot(P1.T, hessian), P1)
        new_hessian = np.dot(np.dot(P2.T, new_hessian), P2)
        new_hessian = np.dot(np.dot(P3.T, new_hessian), P3)
        return new_hessian

    def _read_till_phrase(self,phrase):
        """Read lines from the current file until a match with phrase is found
           Once a match is found return the matching line"""
        line = ""
        while not phrase.match(line):
            line = self.file_descriptor.readline()
        return line

    def _dynamical_matrix(self, hessian):
        """Process the dynamical matrix
        Hessian is a nxn matrix of the mass weighted force constants
        The hessian is symmetrised
        Translational modes are projected out
        The hessian is diagonalised
        Finally the frequencies and normal modes are stored"""
        if self.debug:
            print("_dynamical_matrix")
            print("hessian", hessian[0:4][0])
        #
        nmodes = self.nions*3
        masses = np.array(self.masses)*amu
        if self.debug:
            print("masses", self.masses, masses)
        if not self.nomass_hessian_has_been_set:
            # symmetrise the hessian and store it for later use
            if self.hessian_symmetrisation == "symm":
                hessian = 0.5 * (hessian + hessian.T)
            else:
                # Crystal only uses the upper triangle to calculate the eigenvalues
                hessian = hessian - np.tril(hessian) + np.triu(hessian).T
            self.nomass_hessian_has_been_set = True
            self.nomass_hessian = self._remove_mass_weighting(hessian,masses)
        if self.debug:
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        hessian = self._modify_mass_weighting(self.nomass_hessian,masses)
        if self.debug:
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        if self.debug:
            print("mass weighted hessian", hessian[0:4][0])
        # Project out the translational modes if requested
        if self.eckart:
            hessian = self.project(hessian)
        if self.debug:
            print("projected hessian", hessian[0:4][0])
        # diagonalise
        eig_val, eig_vec = np.linalg.eigh(hessian)
        #
        # If eig_val has negative values then we store the negative frequency
        # convert to cm-1
        frequencies_a = np.zeros(nmodes)
        for i, eig in enumerate(eig_val):
            if eig < 0:
                frequencies_a[i] = -math.sqrt(-eig) / wavenumber
            else:
                frequencies_a[i] = math.sqrt(eig) / wavenumber
            # end if
        # end for
        self.mass_weighted_normal_modes = []
        self.frequencies = frequencies_a.tolist()
        if self.debug:
            print("frequencies", self.frequencies)
        # Store the mass weighted normal modes
        for i in range(nmodes):
            mode = []
            n = 0
            for j in range(self.nions):
                modea = [eig_vec[n][i], eig_vec[n+1][i], eig_vec[n+2][i]]
                n = n + 3
                mode.append(modea)
            self.mass_weighted_normal_modes.append(mode)
        # end for i
        if self.debug:
            print("non mass weighted hessian", self.nomass_hessian[0:4][0])
        return

    def reset_born_charges(self):
        if not self.original_born_charges_are_being_used:
            self.born_charges = self.original_born_charges

    def neutralise_born_charges(self):
        if self.original_born_charges_are_being_used:
            self.original_born_charges = self.born_charges
            self.original_born_charges_are_being_used = False
        self._born_charge_sum_rule()
        return

    def _born_charge_sum_rule(self):
        """Apply a simple charge sum rule to all the elements of the born matrices"""
        total = np.zeros((3, 3))
        born_charges = np.array(self.born_charges)
        new_born_charges = np.zeros_like(self.born_charges)
        total = np.sum(born_charges) / self.nions
        if self.debug:
            print('born charge sum', total)
        new_born_charges = born_charges - total
        self.born_charges = new_born_charges.tolist()
        return

    def _modify_mass_weighting(self,hessian,new):
        #
        # Mass weight defined in new
        #
        ipos = -1
        new_hessian = np.empty_like(hessian)
        for i in range(self.nions):
            for ix in range(3):
                ipos += 1
                jpos = -1
                for j in range(self.nions):
                    for jx in range(3):
                        jpos += 1
                        new_hessian[ipos,jpos] = hessian[ipos,jpos] / math.sqrt(  new[i]*new[j] )
                    # end for jx
                # end for j
            # end for ix
        # end for i
        return new_hessian

    def _remove_mass_weighting(self,hessian,old):
        #
        # Remove the mass weighting imposed by the QM/MM program - defined in old
        #
        new_hessian = np.empty_like(hessian)
        ipos = -1
        for i in range(self.nions):
            for ix in range(3):
                ipos += 1
                jpos = -1
                for j in range(self.nions):
                    for jx in range(3):
                        jpos += 1
                        new_hessian[ipos,jpos] = hessian[ipos,jpos] * math.sqrt( old[i]*old[j] )
                    # end for jx
                # end for j
            # end for ix
        # end for i
        return new_hessian



#!/usr/bin/python
#
# Copyright 2015 John Kendrick
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License
# along with this program, if not see https://opensource.org/licenses/MIT
#
"""Read the contents of a Castep output file
   Inherit the following from the GenericOutputReader
   __init__
   printInfo
   _read_output_file
   _DyanmicalMatrix
"""
import re
import os
import numpy as np
from PDielec.UnitCell import UnitCell
from PDielec.GenericOutputReader import GenericOutputReader


class CastepOutputReader(GenericOutputReader):
    """Read the contents of a Castep output file
       Inherit the following from the GenericOutputReader
       __init__
       printInfo
       _read_output_file
       _DyanmicalMatrix
       _read_till_phrase
"""

    def __init__(self, filenames):
        GenericOutputReader.__init__(self, filenames)
        if filenames[0].find(".castep"):
            seedname, ext = os.path.splitext(filenames[0])
        elif filenames[0].find(".phonon"):
            seedname, ext = os.path.splitext(filenames[0])
        self._castepfile             = seedname+".castep"
        self._phononfile             = seedname+".phonon"
        self.names                   = [self._castepfile, self._phononfile]
        self._outputfiles            = [self._castepfile, self._phononfile]
        self.type                    = 'Castep output'
        # Specific Castep Reader Variables
        self._pspots                  = {}
        self._ediff                   = 0.0
        self._epsilon                 = None
        self._nbranches              = 0
        self._pulay                  = None
        self._ion_type_index         = {}
        self._ion_index_type         = {}
        self._intensities             = None
        return

    def _read_output_files(self):
        """For the .castep file"""
        self.manage = {}   # Empty the dictionary matching phrases
        self.manage['spin']          = (re.compile(' *net spin   of'), self._read_spin)
        self.manage['nelect']        = (re.compile(' *number of  electrons'), self._read_nelect)
        self.manage['cellcontents']  = (re.compile(' *Unit Cell'), self._read_cellcontents)
        self.manage['pspots']        = (re.compile(' *Files used for pseudopotentials:'), self._read_pspot)
        self.manage['masses']        = (re.compile(' *Mass of species in AMU'), self._read_masses)
        self.manage['kpoints']       = (re.compile(' *Number of kpoints used'), self._read_kpoints)
        self.manage['kpoint_grid']   = (re.compile(' *MP grid size for SCF'), self._read_kpoint_grid)
        self.manage['finalenergy']   = (re.compile(' *Final energy, E'), self._read_energies)
        self.manage['finalenergy2']   = (re.compile('Final energy ='), self._read_energies2)
        self.manage['finalenergy3']   = (re.compile('Dispersion corrected final energy'), self._read_energies3)
        self.manage['energy_cutoff'] = (re.compile(' *plane wave basis set cut'), self._read_energy_cutoff)
        self.manage['nbands']        = (re.compile(' *number of bands'), self._read_nbands)
        self.manage['pressure']      = (re.compile(' *\* *Pressure: '), self._read_external_pressure)
        self.manage['opticalDielectric']  = (re.compile(' *Optical Permittivity'), self._read_dielectric)
        self.manage['bornCharges']    = (re.compile(' *Born Effective Charges'), self._read_born_charges)
        #  For the .phonon file
        self.manage['frequency']      = (re.compile('     q-pt=    1    0.000000  0.000000  0.000000      1.0000000000 *$'), self._read_frequencies)
        self.manage['nbranches']      = (re.compile(' Number of branches'), self._read_nbranches)
        for f in self._outputfiles:
            self._read_output_file(f)
        return

    def _read_nbranches(self, line):
        # phonon file being read
        self._nbranches = int(line.split()[3])

    def _read_frequencies(self, line):
        # phonon file being read
        self.mass_weighted_normal_modes = []
        # maybe should use _nbranches
        frequencies = []
        intensities = []
        normal_modes = []
        for imode in range(self._nbranches):
            line = self.file_descriptor.readline()
            frequencies.append(float(line.split()[1]))
            intensities.append(float(line.split()[2]))

        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        for imode in range(self._nbranches):
            a = []
            for ion in range(self.nions):
                line = self.file_descriptor.readline()
                a.append([float(line.split()[2]), float(line.split()[4]), float(line.split()[6])])
            # end for ion
            normal_modes.append(a)
        # end for imode
        self.frequencies = []
        self._intensities = []
        self.mass_weighted_normal_modes = []
        # now reads all frequencies imaginary or not
        # imaginary frequencies are indicated by real negative values
        for i, freq in enumerate(frequencies):
            self.frequencies.append(freq)
            self._intensities.append(intensities[i])
            self.mass_weighted_normal_modes.append(normal_modes[i])
            # end of if freq
        # end of for freq
        return

    def _read_kpoint_grid(self, line):
        self.kpoint_grid = [ int(line.split()[7]), int(line.split()[8]), int(line.split()[9]) ]
        return

    def _read_kpoints(self, line):
        self.kpoints = int(line.split()[5])
        return

    def _read_nbands(self, line):
        self.nbands = int(line.split()[4])
        return

    def _read_pseudoatom(self, line):
        species = line.split()[5].capitalize()
        # These are two private dictionary to map the species name to a type index
        self.species.append(species)
        self._ion_type_index[species] = self.nspecies
        self._ion_index_type[self.nspecies] = species
        self.nspecies += 1
        return

    def _read_cellcontents(self, line):
        line = self._read_till_phrase(re.compile(' *Real Lattice'))
        line = self.file_descriptor.readline()
        avector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        bvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        line = self.file_descriptor.readline()
        cvector = [float(line.split()[0]), float(line.split()[1]), float(line.split()[2])]
        self.unit_cells.append(UnitCell(avector, bvector, cvector))
        self.ncells = len(self.unit_cells)
        line = self._read_till_phrase(re.compile(' *Current cell volume'))
        self.volume = float(line.split()[4])
        if self.nions == 0:
            line = self._read_till_phrase(re.compile(' *Total number of ions in cell'))
            if len(self.unit_cells) == 1:
                self.nions = int(line.split()[7])
                line = self.file_descriptor.readline()
                self.nspecies = int(line.split()[7])
                line = self.file_descriptor.readline()
                line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
        # end if self.nions
        else:
            # Not all Unit Cell output is the same in Castep 17
            line = self._read_till_phrase(re.compile(' *xxxxxxxxxxxxxxxxxxxxxxxxxxxx'))
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
            line = self.file_descriptor.readline()
        # end of else
        fractional_coordinates = []
        # ions_per_type is a dictionary here and is local
        ions_per_type = {}
        self.atom_type_list = []
        self.species = []
        species_list = []
        for i in range(self.nions):
            line = self.file_descriptor.readline()
            atom_type = line.split()[1].capitalize()
            species_list.append(atom_type)
            if atom_type not in ions_per_type:
                self.species.append(atom_type)
                ions_per_type[atom_type] = 0
            self.atom_type_list.append(self.species.index(atom_type))
            ions_per_type[atom_type] += 1
            atom_frac = [float(f) for f in line.split()[3:6]]
            fractional_coordinates.append(atom_frac)
        # At some point we should store the fractional coordinates
        for species in self.species:
            n = ions_per_type[species]
            # self.ions_per_type is a list
            self.ions_per_type.append(n)
        self.unit_cells[-1].set_fractional_coordinates(fractional_coordinates)
        self.unit_cells[-1].set_element_names(species_list)
        return

    def _read_masses(self, line):
        self.masses = []
        self.masses_per_type = []
        self.species = []
        self.nspecies = 0
        line = self.file_descriptor.readline()
        while len(line.split()) != 0:
            species = line.split()[0].capitalize()
            mass = float(line.split()[1])
            nions = self.ions_per_type[self.nspecies]
            self.species.append(species)
            # These are two private dictionary to map the species name to a type index
            # self._ion_type_index[species] = self.nspecies
            self._ion_index_type[self.nspecies] = species
            self.masses_per_type.append(mass)
            self.nspecies += 1
            for j in range(nions):
                self.masses.append(mass)
            line = self.file_descriptor.readline()
        # end while loop
        return

    def _read_born_charges(self, line):
        """Read the born charges from the castep file.
           Each column of the output refers to a given field direction
           Each row in the row refers the atomic displacement
           So the numbers in the output are arrange  [[a1x a2x a3x]
                                                      [a1y a2y a3y]
                                                      [a1z a2z a3z]]
           The output tensor needs them arranged     [[a1x a1y a1z]
                                                      [a2x a2y a2z]
                                                      [a3x a3y a3z]]
           where 1,2,3 are the field directions and x, y, z are the atomic displacements"""
        line = self.file_descriptor.readline()
        self.born_charges = []
        for i in range(self.nions):
            b = []
            line = self.file_descriptor.readline()
            b.append([float(f) for f in line.split()[2:5]])
            line = self.file_descriptor.readline()
            b.append([float(f) for f in line.split()[0:3]])
            line = self.file_descriptor.readline()
            b.append([float(f) for f in line.split()[0:3]])
            B = np.array(b)
            C = B.T
            self.born_charges.append(C.tolist())
        return

    def _read_dielectric(self, line):
        line = self.file_descriptor.readline()
        line = self.file_descriptor.readline()
        # this is epsilon infinity
        self.zerof_optical_dielectric = []
        # this is the zero frequency static dielectric constant
        self.zerof_static_dielectric = []
        self.zerof_optical_dielectric.append([float(f) for f in line.split()[0:3]])
        self.zerof_static_dielectric.append([float(f) for f in line.split()[3:6]])
        line = self.file_descriptor.readline()
        self.zerof_optical_dielectric.append([float(f) for f in line.split()[0:3]])
        self.zerof_static_dielectric.append([float(f) for f in line.split()[3:6]])
        line = self.file_descriptor.readline()
        self.zerof_optical_dielectric.append([float(f) for f in line.split()[0:3]])
        self.zerof_static_dielectric.append([float(f) for f in line.split()[3:6]])
        return

    def _read_skip4(self, line):
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        self.file_descriptor.readline()
        return

    def _read_external_pressure(self, line):
        self.pressure = float(line.split()[2])
        return

    def _read_pspot(self, line):
        for i in range(self.nspecies):
            line = self.file_descriptor.readline()
            self._pspots[line.split()[0]] = line.split()[1]
        return

    def _read_spin(self, line):
        self.spin = int(float(line.split()[5]))
        return

    def _read_energy_cutoff(self, line):
        self.energy_cutoff = float(line.split()[6])
        return

    def _read_ediff(self, line):
        self._ediff = float(line.split()[2])
        return

    def _read_nelect(self, line):
        self.electrons = int(float(line.split()[4]))
        return

    def _read_epsilon(self, line):
        var = line.split()[2].lower()
        self.epsilon = False
        if var == '.true.' or var == 't' or var == 'true':
            self.epsilon = True
        return

    def _read_energies(self, line):
        self.final_energy_without_entropy = float(line.split()[4])
        line = self.file_descriptor.readline()
        self.final_free_energy = float(line.split()[5])
        return

    def _read_energies2(self, line):
        self.final_energy_without_entropy = float(line.split()[3])
        self.final_free_energy = float(line.split()[3])
        return

    def _read_energies3(self, line):
        self.final_energy_without_entropy = float(line.split()[5])
        self.final_free_energy = float(line.split()[5])
        return




#!/usr/bin/python
#
# Copyright 2015 John Kendrick
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License
# along with this program, if not see https://opensource.org/licenses/MIT
#
"""Read the contents of a directory containg Phonopy input and output files"""
import re
import os
import math
import numpy as np
from PDielec.UnitCell import UnitCell
from PDielec.GenericOutputReader import GenericOutputReader
from PDielec.Constants import wavenumber


class PhonopyOutputReader(GenericOutputReader):
    """Read the contents of a directory containg Phonopy input and output files"""

    def __init__(self, names, qmreader):
        GenericOutputReader.__init__(self, names)
        # We have to use the qm reader to do the reading of the QM files
        self.type                    = 'Phonopy output'
        self.qmreader                = qmreader
        return

    def _read_output_files(self):
        """Read the Phonopy files in the directory"""
        # Set the qm reader to have all the settings that phonopy reader has
        self.qmreader.eckart = self.eckart
        self.qmreader.hessian_symmetrisation = self.hessian_symmetrisation
        self.qmreader.debug = self.debug
        # trigger the reading of the qm files
        self.qmreader.read_output()
        # We don't call self._read_outputfile as this starts looking for keywords
        # for f in self._outputfiles:
        #     self._read_output_file(f)
        #
        # Instead copy anything useful from the QM calcs to self
        self.ncells                  = self.qmreader.ncells
        self.unit_cells              = self.qmreader.unit_cells
        self.volume                  = self.qmreader.volume
        self.spin                    = self.qmreader.spin
        self.energy_cutoff           = self.qmreader.energy_cutoff
        self.kpoints                 = self.qmreader.kpoints
        self.kpoint_grid             = self.qmreader.kpoint_grid
        self.nbands                  = self.qmreader.nbands
        self.nions                   = self.qmreader.nions
        self.ions_per_type           = self.qmreader.ions_per_type
        self.atom_type_list          = self.qmreader.atom_type_list
        self.electrons               = self.qmreader.electrons
        self.magnetization           = self.qmreader.magnetization
        self.final_energy_without_entropy = self.qmreader.final_energy_without_entropy
        self.final_free_energy       = self.qmreader.final_free_energy
        self.pressure                = self.qmreader.pressure
        self.masses_per_type         = self.qmreader.masses_per_type
        self.ions_per_type           = self.qmreader.ions_per_type
        self.masses                  = self.qmreader.masses
        self.nspecies                = self.qmreader.nspecies
        self.species                 = self.qmreader.getSpecies()
        self.born_charges            = self.qmreader.born_charges
        self.zerof_optical_dielectric= self.qmreader.zerof_optical_dielectric
        self.zerof_static_dielectric = self.qmreader.zerof_static_dielectric
        # Calculate dynamical matrix
        self.read_dynamical_matrix()
        return

    def read_dynamical_matrix(self):
        #
        # Yaml imports of large files are really slow....
        # Attempt to use the PyYaml C parser, using yaml.CLoader
        #
        import yaml
        try:
            from yaml import CLoader as Loader
        except:
            print("WARNING: Yaml CLoader is not avaiable, using fallback")
            from yaml import Loader as Loader
        # the first name has to be the qpoints file
        fd = open(self._outputfiles[0])
        data_q = yaml.load(fd, Loader=Loader)
        fd.close
        # the second name has to be the phonopy file
        fd = open(self._outputfiles[1])
        data_p = yaml.load(fd, Loader=Loader)
        fd.close
        self._old_masses = []
        for i in range(self.nions):
            self._old_masses.append(data_p['primitive_cell']['points'][i]['mass'])
        qpoints = data_q['phonon'][0]['q-position']
        # print('q-points',qpoints)
        natom = data_q['natom']
        # print('natom:',natom)
        dynmat = []
        dynmat_data = data_q['phonon'][0]['dynamical_matrix']
        for row in dynmat_data:
            vals = np.reshape(row, (-1, 2))
            dynmat.append(vals[:, 0] + vals[:, 1] * 1j)
        dynmat = np.array(dynmat)
        # Make sure the hessian is real
        hessian = np.real(dynmat)
        # We need to convert to cm-1
        conversion_factor_to_THz = 15.633302
        conversion_factor_to_cm1 = conversion_factor_to_THz * 33.35641
        conv  = conversion_factor_to_cm1
        hessian = hessian * conv * conv
        # Find its eigenvalues and eigen vectors
        eig_val, eig_vec = np.linalg.eigh(hessian)
        self.mass_weighted_normal_modes = []
        nmodes = 3*self.nions
        # Store the new frequencies, using the negative convention for imaginary modes
        frequencies_a = np.sqrt(np.abs(eig_val.real)) * np.sign(eig_val.real)
        self.frequencies = frequencies_a.tolist()
        # Store the mass weighted normal modes
        for i in range(nmodes):
            mode = []
            n = 0
            for j in range(self.nions):
                modea = [eig_vec[n][i], eig_vec[n+1][i], eig_vec[n+2][i]]
                n = n + 3
                mode.append(modea)
            self.mass_weighted_normal_modes.append(mode)
        # end for i
        return




#!/usr/bin/python
"""Do some plotting for the dielectric functions"""
from __future__ import print_function, division
import sys
import numpy as np


class Plotter:
    """Do some plotting for the dielectric functions"""
    def __init__(self):
        self.methods = []
        self.volume_fractions = []
        self.sizes = []
        self.size_sigmas = []
        self.shapes = []
        self.shape_data = []
        self.plot_numbers = []
        self.traces = []
        self.absorption_coefficients = []
        self.molar_absorption_coefficients = []
        self.frequencies = []
        return

    def add_dielectric(self, nplot, method, vf_type, size, size_sigma, shape, data, frequency, trace, absorption, molar_absorption):
        """Add the trace of a dielectric to be plotted, add the absorption and the molar absorption"""
        if nplot not in self.plot_numbers:
            self.plot_numbers.append(nplot)
            self.methods.append(method)
            self.volume_fractions.append(vf_type)
            self.sizes.append(size)
            self.size_sigmas.append(size_sigma)
            self.shapes.append(shape)
            self.shape_data.append(data)
            self.traces.append([])
            self.absorption_coefficients.append([])
            self.molar_absorption_coefficients.append([])
            self.frequencies.append([])
        self.frequencies[nplot].append(frequency)
        self.traces[nplot].append(trace)
        self.absorption_coefficients[nplot].append(absorption)
        self.molar_absorption_coefficients[nplot].append(molar_absorption)
        return

    def excel(self, workbook):
        """Print data to a xlsx file"""
        import xlsxwriter as xlsx
        #  Loop over the different sheets
        worksheets = []
        for itype,name in enumerate(['molar_absorption_coefficient','real_dielectric','imag_dielectric','absorption_coefficent']):
          worksheet = workbook.add_worksheet(name)
          worksheets.append(worksheet)
          # Write a header
          if itype == 0:
            worksheet.write(0,0,'Molar Absorption Coefficients in cm-1.L.mol-1')
          elif itype == 1:
            worksheet.write(0,0,'Real component of the permittivity')
          elif itype == 2:
            worksheet.write(0,0,'Imaginary component of the permittivity')
          elif itype == 3:
            worksheet.write(0,0,'Absorption coefficient')
          col = 1
          # loop over the settings for each set of data
          for method, shape, data, vf_type, size, size_sigma in zip(self.methods, self.shapes, self.shape_data, self.volume_fractions, self.sizes, self.size_sigmas):
            col += 1;
            row = 0
            row += 1; worksheet.write(row,0, 'Method'); worksheet.write(row,col, method)
            row += 1; worksheet.write(row,0, 'Shape'); worksheet.write(row,col, shape)
            row += 1; worksheet.write(row,0, 'Shape Data'); worksheet.write(row,col, data)
            row += 1; worksheet.write(row,0, 'Mass or Volume fraction'); worksheet.write(row,col, vf_type)
            row += 1; worksheet.write(row,0, 'Particle size um'); worksheet.write(row,col, size)
            row += 1; worksheet.write(row,0, 'Sigma of size distribution'); worksheet.write(row,col, size_sigma)
            strdata = str(data)
            strdata = strdata.replace(" ", "")
            strdata = strdata.replace(",", " ")
            strdata = strdata.replace("'", "")
            strdata = strdata.replace('"', "")
            ssize=""
            if size > 1.0e-6:
                ssize = "/{:f}".format(size).rstrip("0").rstrip(".")
                ssize += "mu"
                if size_sigma:
                    ssize =  "/{:f}".format(size).rstrip("0").rstrip(".")
                    ssize += "/{:f}".format(size_sigma).rstrip("0").rstrip(".")
                    ssize += "mu"
            row += 1
            worksheet.write(row,0, 'cm-1')
            worksheet.write(row,1, 'THz')
            if strdata == "":
                worksheet.write(row,col,method+"/"+shape+"/"+name+"("+vf_type+ssize+")")
            else:
                worksheet.write(row,col,method+"/"+shape+"/"+strdata+"/"+name+"("+vf_type+ssize+")")
            lastrow = row
          # end loop over method.....
        # end of loop over worksheets
        # Use the first frequency list and write the frequencies in cm-1 and THz in each worksheet
        for worksheet in worksheets:
            col = 0
            row = lastrow
            for v in self.frequencies[0]:
                thz = v * 0.0299792458
                row += 1
                worksheet.write(row,col,v)
                worksheet.write(row,col+1,thz)
        # Write entries in each worksheet
        col = 1
        for molar_absorption in self.molar_absorption_coefficients:
            col += 1
            row = lastrow
            for entry in molar_absorption:
                row += 1
                worksheets[0].write(row,col,entry)
            # end loop molar absorption
        col = 1
        for traces in np.real(self.traces):
            col += 1
            row = lastrow
            for entry in traces:
                row += 1
                worksheets[1].write(row,col,entry)
            # end loop dri
        col = 1
        for traces in np.imag(self.traces):
            col += 1
            row = lastrow
            for entry in traces:
                row += 1
                worksheets[2].write(row,col,entry)
            # end loop dri
        col = 1
        for absorption in self.absorption_coefficients:
            col += 1
            row = lastrow
            for entry in absorption:
                row += 1
                worksheets[3].write(row,col,entry)
            # end loop absorption
        # end loop over frequency
        return
    def printout(self, fd_csvfile, csv_extended):
        """Print data to a csv file"""
        if fd_csvfile == 0:
            return
        # Print the header
        output = ",,,,,,cm-1,thz"
        if csv_extended:
          output = "cm-1,thz"
        for outputType in ['_molar_absorption_coefficient_cm-1.L.mole-1','_real_dielectric','_imag_dielectric','_absorption_coefficent_cm-1']:
          for method, shape, data, vf_type, size, size_sigma in zip(self.methods, self.shapes, self.shape_data, self.volume_fractions, self.sizes, self.size_sigmas):
            strdata = str(data)
            strdata = strdata.replace(" ", "")
            strdata = strdata.replace(",", " ")
            strdata = strdata.replace("'", "")
            strdata = strdata.replace('"', "")
            ssize=""
            if size > 1.0e-6:
                ssize = "/{:f}".format(size).rstrip("0").rstrip(".")
                ssize += "mu"
                if size_sigma:
                    ssize =  "/{:f}".format(size).rstrip("0").rstrip(".")
                    ssize += "/{:f}".format(size_sigma).rstrip("0").rstrip(".")
                    ssize += "mu"
            output += ","+method+"_"+shape+strdata+outputType+"("+vf_type+ssize+")"
          # end loop over method.....
        # end of loop over outputType
        print(output, file=fd_csvfile)
        # Use the first frequency list
        for iv, v in enumerate(self.frequencies[0]):
            thz = v * 0.0299792458
            output = ",,,,,,{:12f},{:12f}".format(v, thz)
            if csv_extended:
                output = "{:12f},{:12f}".format(v, thz)
            for molar_absorption in self.molar_absorption_coefficients:
                output += ",{:20.8f}".format(molar_absorption[iv])
                # end loop dri, absorption
            for dri in self.traces:
                trace_real = float(np.real(dri[iv]))
                output += ",{:20.8f}".format(trace_real)
            # end loop dri
            for dri in self.traces:
                trace_imag = float(np.imag(dri[iv]))
                output += ",{:20.8f}".format(trace_imag)
                # end loop dri
            for absorption in self.absorption_coefficients:
                output += ",{:20.8f}".format(absorption[iv])
                # end loop absorption
            print(output, file=fd_csvfile)
        # end loop over frequency
        return

    def plot(self, types):
        """Plot driver routine"""
        for plot in types:
            if plot == 'imaginary':
                self.plot_imaginary()
            elif plot == 'real':
                self.plot_real()
            elif plot == 'extinction':
                self.plot_absorption()
            elif plot == 'absorption':
                self.plot_absorption()
            elif plot == 'molar_extinction':
                self.plot_molar_absorption()
            elif plot == 'molarExtinction':
                self.plot_molar_absorption()
            elif plot == 'molarextinction':
                self.plot_molar_absorption()
            elif plot == 'molar_absorption':
                self.plot_molar_absorption()
            elif plot == 'molarAbsorption':
                self.plot_molar_absorption()
            elif plot == 'molarabsorption':
                self.plot_molar_absorption()
            else:
                print('Unkown plot type {}'.format(plot))
                exit(1)
            # endif
        # endfor
        return

    def plot_molar_absorption(self):
        """Plot loss data """
        import pylab as pl
        # Choose a suitable scale TO has to be treated differently, TO was added last
        maxd = 0.0
        max_to = 1.2*np.amax(self.molar_absorption_coefficients[-1])
        # loop over all methods except the last, TO
        for molar_absorption in self.molar_absorption_coefficients[:-1]:
            maxd = max(maxd, np.amax(molar_absorption))
        # end loop over entries
        to_scale = maxd/max_to
        to_scale = 1.0
        molar_absorption = np.array(self.molar_absorption_coefficients[-1])
        self.molar_absorption_coefficients[-1] = to_scale * molar_absorption
        plot_list = []
        plot_labels = []
        for method, shape, data, frequencies, molar_absorption, vf_type, size, size_sigma in zip(self.methods, self.shapes, self.shape_data, self.frequencies, self.molar_absorption_coefficients, self.volume_fractions, self.sizes, self.size_sigmas):
            ssize=""
            if size > 1.0e-6:
                ssize = "/{:f}".format(size).rstrip("0").rstrip(".")
                ssize += "mu"
                if size_sigma:
                    ssize =  "/{:f}".format(size).rstrip("0").rstrip(".")
                    ssize += "/{:f}".format(size_sigma).rstrip("0").rstrip(".")
                    ssize += "mu"
            label = method+"_"+shape+str(data)+"("+vf_type+ssize+")"
            label = label.replace(" ", "")
            label = label.replace(",", " ")
            label = label.replace("'", "")
            label = label.replace('"', '')
            plot, = pl.plot(frequencies, molar_absorption, label=label)
            plot_list.append(plot)
            plot_labels.append(label)
        # end loop over method
        pl.legend(plot_list,  plot_labels, loc='best', numpoints=1)
        pl.xlabel('Frequency (cm-1)')
        pl.ylabel('Molar Absorption (cm-1.L.mole-1)')
        pl.title('Molar Absorption')
        pl.show()
        return

    def plot_absorption(self):
        """Plot absorption data """
        import pylab as pl
        # Choose a suitable scale TO has to be treated differently, TO was added last
        maxd = 0.0
        max_to = 1.2*np.amax(self.absorption_coefficients[-1])
        # loop over all methods except the last, TO
        for absorptions in self.absorption_coefficients[:-1]:
            maxd = max(maxd, np.amax(absorptions))
        # end loop over entries
        to_scale = maxd/max_to
        to_scale = 1.0
        absorptions = np.array(self.absorption_coefficients[-1])
        self.absorption_coefficients[-1] = to_scale * absorptions
        plot_list = []
        plot_labels = []
        for method, shape, data, frequencies, absorptions, vf_type, size, size_sigma in zip(self.methods, self.shapes, self.shape_data, self.frequencies, self.absorption_coefficients, self.volume_fractions, self.sizes, self.size_sigmas):
            ssize=""
            if size > 1.0e-6:
                ssize = "/{:f}".format(size).rstrip("0").rstrip(".")
                ssize += "mu"
                if size_sigma:
                    ssize =  "/{:f}".format(size).rstrip("0").rstrip(".")
                    ssize += "/{:f}".format(size_sigma).rstrip("0").rstrip(".")
                    ssize += "mu"
            label = method+"_"+shape+str(data)+"("+vf_type+ssize+")"
            label = label.replace(" ", "")
            label = label.replace(",", " ")
            label = label.replace("'", "")
            label = label.replace('"', '')
            plot, = pl.plot(frequencies, absorptions, label=label)
            plot_list.append(plot)
            plot_labels.append(label)
        # end loop over method
        pl.legend(plot_list,  plot_labels, loc='best', numpoints=1)
        pl.xlabel('Frequency (cm-1)')
        pl.ylabel('Absorption Coefficient (cm-1)')
        pl.title('Absorption Coefficient')
        pl.show()
        return

    def plot_imaginary(self):
        """Plot loss data """
        import pylab as pl
        # Choose a suitable scale TO has to be treated differently, TO was added last
        maxd = 0.0
        max_to = 1.2*np.amax(np.imag(self.traces[-1]))
        # loop over all methods except the last, TO
        for traces in self.traces[:-1]:
            real_trace = np.imag(traces)
            maxd = max(maxd, np.amax(real_trace))
        # end loop over entries
        to_scale = maxd/max_to
        to_scale = 1.0
        traces = np.array(self.traces[-1])
        self.traces[-1] = to_scale * traces
        plot_list = []
        plot_labels = []
        for method, shape, data, frequencies, traces, vf_type, size, size_sigma in zip(self.methods, self.shapes, self.shape_data, self.frequencies, self.traces, self.volume_fractions, self.sizes, self.size_sigmas):
            ssize=""
            if size > 1.0e-6:
                ssize = "/{:f}".format(size).rstrip("0").rstrip(".")
                ssize += "mu"
                if size_sigma:
                    ssize =  "/{:f}".format(size).rstrip("0").rstrip(".")
                    ssize += "/{:f}".format(size_sigma).rstrip("0").rstrip(".")
                    ssize += "mu"
            label = method+"_"+shape+str(data)+"("+vf_type+ssize+")"
            label = label.replace(" ", "")
            label = label.replace(",", " ")
            label = label.replace("'", "")
            label = label.replace('"', '')
            plot, = pl.plot(frequencies, np.imag(traces), label=label)
            plot_list.append(plot)
            plot_labels.append(label)
        # end loop over method
        pl.legend(plot_list,  plot_labels, loc='best', numpoints=1)
        pl.xlabel('Frequency (cm-1)')
        pl.ylabel('Imaginary Permittivity')
        pl.title('Imaginary Permittivity')
        pl.show()
        return

    def plot_real(self):
        """Plot real data """
        import pylab as pl
        # Choose a suitable scale TO has to be treated differently, TO was added last
        maxd = 0.0
        max_to = 1.2*np.amax(np.real(self.traces[-1]))
        # loop over all methods except the last, TO
        for traces in self.traces[:-1]:
            real_trace = np.real(traces)
            maxd = max(maxd, np.amax(real_trace))
        # end loop over entries
        to_scale = maxd/max_to
        to_scale = 1.0
        traces = np.array(self.traces[-1])
        self.traces[-1] = to_scale * traces
        plot_list = []
        plot_labels = []
        for method, shape, data, frequencies, traces, vf_type, size, size_sigma in zip(self.methods, self.shapes, self.shape_data, self.frequencies, self.traces, self.volume_fractions, self.sizes, self.size_sigmas):
            ssize=""
            if size > 1.0e-6:
                ssize = "/{:f}".format(size).rstrip("0").rstrip(".")
                ssize += "mu"
                if size_sigma:
                    ssize =  "/{:f}".format(size).rstrip("0").rstrip(".")
                    ssize += "/{:f}".format(size_sigma).rstrip("0").rstrip(".")
                    ssize += "mu"
            label = method+"_"+shape+str(data)+"("+vf_type+ssize+")"
            label = label.replace(" ", "")
            label = label.replace(",", " ")
            label = label.replace("'", "")
            label = label.replace('"', '')
            plot, = pl.plot(frequencies, np.real(traces), label=label)
            plot_list.append(plot)
            plot_labels.append(label)
        # end loop over method
        pl.legend(plot_list,  plot_labels, loc='best', numpoints=1)
        pl.xlabel('Frequency (cm-1)')
        pl.ylabel('Real Permittivity')
        pl.title('Real Permittivity')
        pl.show()
        return


def print_ints(title, ints, no_per_line=8, format="{:9d}", file=sys.stdout, separator=" "):
    """Print ints data """
    #
    # Print out a list of ints prettily
    #
    len_ints = len(ints)
    if title != "":
        print(" ", file=file)
        print(title, file=file)
    nlines = int((len_ints - 1) / no_per_line) + 1
    start = 0
    for i in range(nlines):
        end = start + no_per_line
        if end > len_ints:
            end = len_ints
        print(" " + separator.join(format.format(r) for r in ints[start:end]), file=file)
        start = start + no_per_line
    # end for i
    return

def print_strings(title, strings, no_per_line=8, format="{:9s}", file=sys.stdout, separator=" "):
    """Print strings data """
    #
    # Print out a list of strings prettily
    #
    len_strings = len(strings)
    if title != "":
        print(" ", file=file)
        print(title, file=file)
    nlines = int((len_strings - 1) / no_per_line) + 1
    start = 0
    for i in range(nlines):
        end = start + no_per_line
        if end > len_strings:
            end = len_strings
        print(" " + separator.join(format.format(r) for r in strings[start:end]), file=file)
        start = start + no_per_line
    # end for i
    return


def print_reals(title, reals, no_per_line=8, format="{:9.2f}", file=sys.stdout, separator=" "):
    """Print reals data """
    #
    # Print out a list of reals prettily
    #
    len_reals = len(reals)
    if title != "":
        print(" ", file=file)
        print(title, file=file)
    nlines = int((len_reals - 1) / no_per_line) + 1
    start = 0
    for i in range(nlines):
        end = start + no_per_line
        if end > len_reals:
            end = len_reals
        print(" " + separator.join(format.format(r) for r in reals[start:end]), file=file)
        start = start + no_per_line
    # end for i
    return


def print3x3(title, array, format="{:14.6f}", file=sys.stdout, separator=" "):
    """Print 3x3 matrix"""
    #
    # Print out a 3x3 tensor matrix
    #
    print(" ", file=file)
    if title != "":
        print(title, file=file)
    for i in range(3):
        print("      "+" ".join(format.format(p) for p in array[i]), file=file)
    # end for i
    return



import sys
import os.path
import os
import numpy as np
osname = os.name
from multiprocessing import cpu_count, Array
import PDielec.Calculator as Calculator
from PyQt5.QtWidgets  import  QPushButton, QWidget
from PyQt5.QtWidgets  import  QComboBox, QLabel, QLineEdit
from PyQt5.QtWidgets  import  QProgressBar, QApplication
from PyQt5.QtWidgets  import  QVBoxLayout, QHBoxLayout, QFormLayout
from PyQt5.QtWidgets  import  QSpinBox,QDoubleSpinBox
from PyQt5.QtWidgets  import  QSizePolicy
from PyQt5.QtCore     import  QCoreApplication, Qt
from PDielec.Constants import  wavenumber, amu, PI, avogadro_si, angstrom
from PDielec.Constants import  average_masses, isotope_masses
import ctypes
# Import plotting requirements
import matplotlib
import matplotlib.figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from PDielec.Utilities import Debug
import time

def set_affinity_on_worker():
    '''When a new worker process is created, the affinity is set to all CPUs'''
    #JK print('I'm the process %d, setting affinity to all CPUs.' % os.getpid())
    #JK Commented out for the time being
    #JK os.system('taskset -p 0xff %d > /dev/null' % os.getpid())


class PlottingTab(QWidget):
    def __init__(self, parent, debug=False ):
        super(QWidget, self).__init__(parent)
        global debugger
        debugger = Debug(debug,'PlottingTab')
        self.dirty = True
        self.settings = {}
        self.subplot = None
        self.setWindowTitle('Plotting')
        self.settings['Minimum frequency'] = 0
        self.settings['Maximum frequency'] = 400
        self.settings['Frequency increment'] = 0.2
        self.molar_definitions = ['Unit cells','Atoms','Molecules']
        self.settings['Molar definition'] = 'Unit cells'
        self.settings['Number of atoms'] = 1
        self.settings['Plot title'] = 'Plot Title'
        self.xaxes = []
        self.directions = []
        self.depolarisations = []
        self.frequency_units = None
        self.molar_cb_current_index = 0
        # store the notebook
        self.notebook = parent
        # get the reader from the main tab
        self.reader = self.notebook.mainTab.reader
        # Create the parallel environment
        if self.notebook.threading:
            from multiprocessing.dummy import Pool
        else:
            from multiprocessing import Pool
        # Create last tab - PlottingTab
        vbox = QVBoxLayout()
        form = QFormLayout()
        #
        # The minimum frequency
        #
        self.vmin_sb = QSpinBox(self)
        self.vmin_sb.setRange(0,9000)
        self.vmin_sb.setValue(self.settings['Minimum frequency'])
        self.vmin_sb.setToolTip('Set the minimum frequency to be considered)')
        self.vmin_sb.valueChanged.connect(self.on_vmin_changed)
        label = QLabel('Minimum frequency:', self)
        label.setToolTip('Set the minimum frequency to be considered)')
        form.addRow(label, self.vmin_sb)
        #
        # The maximum frequency
        #
        self.vmax_sb = QSpinBox(self)
        self.vmax_sb.setRange(0,9000)
        self.vmax_sb.setValue(self.settings['Maximum frequency'])
        self.vmax_sb.setToolTip('Set the maximum frequency to be considered)')
        self.vmax_sb.valueChanged.connect(self.on_vmax_changed)
        label = QLabel('Maximum frequency:', self)
        label.setToolTip('Set the maximum frequency to be considered)')
        form.addRow(label, self.vmax_sb)
        #
        # Choose a suitable increment
        #
        self.vinc_sb = QDoubleSpinBox(self)
        self.vinc_sb.setRange(0.0001,5.0)
        self.vinc_sb.setSingleStep(0.1)
        self.vinc_sb.setDecimals(4)
        self.vinc_sb.setToolTip('Choose an increment for the frequency when plotting')
        self.vinc_sb.setValue(self.settings['Frequency increment'])
        self.vinc_sb.valueChanged.connect(self.on_vinc_changed)
        label = QLabel('Frequency increment', self)
        label.setToolTip('Choose an increment for the frequency when plotting')
        form.addRow(label, self.vinc_sb)
        #
        # Define molar quantity
        #
        self.molar_cb = QComboBox(self)
        self.molar_cb.setToolTip('Define what a mole is.  \nIn the case of Molecules, the number of atoms in a molecule must be given')
        self.molar_cb.addItems(self.molar_definitions)
        try:
            self.molar_cb_current_index = self.molar_definitions.index(self.settings["Molar definition"])
        except:
            self.molar_cb_current_index = 0
            self.settings["Molar definition"] = self.molar_definitions[self.molar_cb_current_index]
        self.molar_cb.setCurrentIndex(self.molar_cb_current_index)
        self.molar_cb.activated.connect(self.on_molar_cb_activated)
        label = QLabel('Molar definition', self)
        label.setToolTip('Define what a mole is.  \nIn the case of Molecules, the number of atoms in a molecule must be given')
        form.addRow(label, self.molar_cb)
        #
        # Number of atoms in a molecule
        #
        self.natoms_sb = QSpinBox(self)
        self.natoms_sb.setToolTip('Set the number of atoms in a molecule. \nOnly need this if moles of molecules is needed')
        self.natoms_sb.setRange(1,500)
        self.natoms_sb.setValue(self.settings['Number of atoms'])
        self.natoms_sb.valueChanged.connect(self.on_natoms_changed)
        self.natoms_sb.setEnabled(False)
        label = QLabel('Number of atoms per molecule', self)
        label.setToolTip('Set the number of atoms in a molecule. \nOnly need this if moles of molecules is needed')
        form.addRow(label, self.natoms_sb)
        #
        # Set the plot title
        #
        self.title_le = QLineEdit(self)
        self.title_le.setToolTip('Set the plot title')
        self.title_le.setText(self.settings['Plot title'])
        self.title_le.textChanged.connect(self.on_title_changed)
        label = QLabel('Plot title', self)
        label.setToolTip('Set the plot title')
        form.addRow(label, self.title_le)
        #
        # Set the x-axis frequency units
        #
        self.funits_cb = QComboBox(self)
        self.funits_cb.setToolTip('Set the frequency units for the x-axis')
        self.funits_cb.addItems( ['wavenumber','THz'] )
        self.frequency_units = 'wavenumber'
        self.funits_cb.activated.connect(self.on_funits_cb_activated)
        label = QLabel('Frequency units for the x-axis', self)
        label.setToolTip('Set the frequency units for the x-axis')
        form.addRow(label, self.funits_cb)
        #
        # Final button
        #
        hbox = QHBoxLayout()
        self.molarAbsorptionButton = QPushButton('Plot molar absorption')
        self.molarAbsorptionButton.setToolTip('Plot the molar absorption using the scenarios')
        self.molarAbsorptionButton.clicked.connect(self.molarAbsorptionButtonClicked)
        hbox.addWidget(self.molarAbsorptionButton)
        self.absorptionButton = QPushButton('Plot absorption')
        self.absorptionButton.setToolTip('Plot the absorption using the scenarios')
        self.absorptionButton.clicked.connect(self.absorptionButtonClicked)
        hbox.addWidget(self.absorptionButton)
        self.realPermButton = QPushButton('Plot real permittivity')
        self.realPermButton.setToolTip('Plot the real permittivity using the scenarios')
        self.realPermButton.clicked.connect(self.realPermButtonClicked)
        hbox.addWidget(self.realPermButton)
        self.imagPermButton = QPushButton('Plot imaginary permittivity')
        self.imagPermButton.setToolTip('Plot the imaginary permittivity using the scenarios')
        self.imagPermButton.clicked.connect(self.imagPermButtonClicked)
        hbox.addWidget(self.imagPermButton)
        self.atrButton = QPushButton('Plot ATR')
        self.atrButton.setToolTip('Plot Reflectance ATR')
        self.atrButton.clicked.connect(self.atrButtonClicked)
        hbox.addWidget(self.atrButton)
        form.addRow(hbox)
        # Add a progress bar
        self.progressbar = QProgressBar(self)
        self.progressbar.setToolTip('Show the progress of any calculations')
        self.progressbar.setMinimum(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setMinimum(0)
        label = QLabel('Calculation progress', self)
        label.setToolTip('Show the progress of any calculations')
        form.addRow(label,self.progressbar)
        # Add the matplotlib figure to the bottom
        self.figure = matplotlib.figure.Figure()
        self.canvas = FigureCanvas(self.figure)
        self.canvas.setSizePolicy(QSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding))
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.toolbar.setSizePolicy(QSizePolicy(QSizePolicy.Fixed,QSizePolicy.Fixed))
        form.addRow(self.canvas)
        form.addRow(self.toolbar)
        vbox.addLayout(form)
        # finalise the layout
        self.setLayout(vbox)
        QCoreApplication.processEvents()

    def molarAbsorptionButtonClicked(self):
        debugger.print('molarAbsorptionButtonClicked pressed')
        if self.notebook.plottingCalculationRequired:
            self.calculate()
        if not self.notebook.plottingCalculationRequired:
            self.plot(self.xaxes, self.molarAbsorptionCoefficients, r'Molar Absorption Coefficient $\mathdefault{(L mole^{-1} cm^{-1})}$')

    def absorptionButtonClicked(self):
        debugger.print('absorptionButtonClicked pressed')
        if self.notebook.plottingCalculationRequired:
            self.calculate()
        if not self.notebook.plottingCalculationRequired:
            self.plot(self.xaxes, self.absorptionCoefficients, r'Absorption Coefficient $\mathdefault{(cm^{-1})}$')

    def realPermButtonClicked(self):
        debugger.print('realPermButtonClicked pressed')
        if self.notebook.plottingCalculationRequired:
            self.calculate()
        if not self.notebook.plottingCalculationRequired:
            self.plot(self.xaxes, self.realPermittivities, 'Real Component of Permittivity')

    def imagPermButtonClicked(self):
        debugger.print('imagPermButtonClicked pressed')
        if self.notebook.plottingCalculationRequired:
            self.calculate()
        if not self.notebook.plottingCalculationRequired:
            self.plot(self.xaxes, self.imagPermittivities, 'Imaginary Component of Permittivity')

    def atrButtonClicked(self):
        debugger.print('atrButtonClicked pressed')
        if self.notebook.plottingCalculationRequired:
            self.calculate()
        if not self.notebook.plottingCalculationRequired:
            self.plot(self.xaxes, self.sp_atrs, 'ATR absorption')

    def on_title_changed(self,text):
        self.settings['Plot title'] = text
        if self.subplot is not None:
            self.subplot.set_title(self.settings['Plot title'])
            self.canvas.draw_idle()
        debugger.print('on title change ', self.settings['Plot title'])

    def on_vinc_changed(self,value):
        self.settings['Frequency increment'] = value
        self.notebook.plottingCalculationRequired = True
        self.dirty = True
        self.progressbar.setValue(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setValue(0)
        QCoreApplication.processEvents()
        debugger.print('on vinc change ', self.settings['Frequency increment'])

    def on_vmin_changed(self):
        self.settings['Minimum frequency'] = self.vmin_sb.value()
        self.notebook.plottingCalculationRequired = True
        self.dirty = True
        self.progressbar.setValue(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setValue(0)
        QCoreApplication.processEvents()
        debugger.print('on vmin change ', self.settings['Minimum frequency'])

    def on_vmax_changed(self):
        self.settings['Maximum frequency'] = self.vmax_sb.value()
        self.notebook.plottingCalculationRequired = True
        self.dirty = True
        self.progressbar.setValue(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setValue(0)
        QCoreApplication.processEvents()
        debugger.print('on vmax change ', self.settings['Maximum frequency'])

    def refresh(self,force=False):
        if not self.dirty and not force and not self.notebook.plottingCalculationRequired:
            debugger.print('refreshing widget aborted', self.dirty,force,self.notebook.plottingCalculationRequired)
            return
        debugger.print('refreshing widget', force)
        #
        # Block signals during refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(True)
        # Now refresh values
        self.vmin_sb.setValue(self.settings['Minimum frequency'])
        self.vmax_sb.setValue(self.settings['Maximum frequency'])
        self.vinc_sb.setValue(self.settings['Frequency increment'])
        try:
            self.molar_cb_current_index = self.molar_definitions.index(self.settings["Molar definition"])
        except:
            self.molar_cb_current_index = 0
            self.settings["Molar definition"] = self.molar_definitions[self.molar_cb_current_index]
        self.molar_cb.setCurrentIndex(self.molar_cb_current_index)
        self.natoms_sb.setValue(self.settings['Number of atoms'])
        self.title_le.setText(self.settings['Plot title'])
        # Refresh the widgets that depend on the reader
        self.reader = self.notebook.reader
        if self.reader is not None:
            self.settings['concentration'] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24)
        # Flag a recalculation will be required
        self.notebook.plottingCalculationRequired = True
        # Reset the progress bar
        self.progressbar.setValue(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setValue(0)
        self.molarAbsorptionButtonClicked()
        self.dirty = False
        self.notebook.plottingCalculationRequired = False
        #
        # Unblock signals after refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(False)
        QCoreApplication.processEvents()
        return

    def on_natoms_changed(self, value):
        self.notebook.plottingCalculationRequired = True
        self.dirty = True
        self.progressbar.setValue(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setValue(0)
        QCoreApplication.processEvents()
        self.settings['Number of atoms'] = value
        debugger.print('on natoms changed ', self.settings['Number of atoms'])

    def on_funits_cb_activated(self, index):
        if index == 0:
            self.frequency_units = 'wavenumber'
        else:
            self.frequency_units = 'THz'
        self.replot()
        debugger.print('Frequency units changed to ', self.frequency_units)

    def on_molar_cb_activated(self, index):
        self.molar_cb_current_index = index
        self.settings['Molar definition'] = self.molar_definitions[index]
        self.notebook.plottingCalculationRequired = True
        self.dirty = True
        self.progressbar.setValue(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setValue(0)
        QCoreApplication.processEvents()
        if self.settings['Molar definition'] == 'Molecules':
            self.settings['concentration'] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24 * self.settings['Number of atoms'] / self.reader.nions)
            self.natoms_sb.setEnabled(True)
        elif self.settings['Molar definition'] == 'Unit cells':
            self.settings['concentration'] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24)
            self.natoms_sb.setEnabled(False)
        elif self.settings['Molar definition'] == 'Atoms':
            self.settings['concentration'] = 1000.0 / (avogadro_si * self.reader.volume * 1.0e-24 / self.reader.nions)
            self.natoms_sb.setEnabled(False)
        debugger.print('The concentration has been set', self.settings['Molar definition'], self.settings['concentration'])

    def calculate(self):
        debugger.print('calculate')
        self.progressbar.setValue(0)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setValue(0)
        QCoreApplication.processEvents()
        # Assemble the mainTab settings
        settings = self.notebook.mainTab.settings
        program = settings['Program']
        filename = settings['Output file name']
        reader = self.notebook.mainTab.reader
        if reader is None:
            return
        if program == '':
            return
        if filename == '':
            return
        QApplication.setOverrideCursor(Qt.WaitCursor)
        # Assemble the settingsTab settings
        self.notebook.plottingCalculationRequired = False
        settings = self.notebook.settingsTab.settings
        eckart = settings['Eckart flag']
        neutral = settings['Neutral Born charges']
        epsilon_inf = np.array(settings['Optical permittivity'])
        sigmas_cm1 = self.notebook.settingsTab.sigmas_cm1
        sigmas = np.array(sigmas_cm1) * wavenumber
        modes_selected = self.notebook.settingsTab.modes_selected
        frequencies_cm1 = self.notebook.settingsTab.frequencies_cm1
        frequencies = np.array(frequencies_cm1) * wavenumber
        intensities = self.notebook.settingsTab.intensities
        oscillator_strengths = self.notebook.settingsTab.oscillator_strengths
        volume = reader.volume*angstrom*angstrom*angstrom
        vmin = self.settings['Minimum frequency']
        vmax = self.settings['Maximum frequency']
        vinc = self.settings['Frequency increment']
        calling_parameters = []
        self.scenarios = self.notebook.scenarios
        mode_list = []
        drude_sigma = 0
        drude_plasma = 0
        drude = False
        for mode_index,selected in enumerate(modes_selected):
            if selected:
                mode_list.append(mode_index)
        #debugger.print('mode_list', mode_list)
        # Calculate the ionic permittivity at zero frequency
        epsilon_ionic = self.notebook.settingsTab.epsilon_ionic
        epsilon_total = epsilon_inf + epsilon_ionic
        cell = reader.unit_cells[-1]
        self.directions = []
        self.depolarisations = []
        # Process the shapes and the unique directions.
        # and calculate the depolarisation matrices
        self.legends = []
        for scenario in self.scenarios:
            self.legends.append(scenario.settings['Legend'])
            #debugger.print('Scenario ',scenario.scenarioIndex)
            shape = scenario.settings['Particle shape']
            #debugger.print('Particle shape ',shape)
            hkl = [scenario.settings['Unique direction - h'], scenario.settings['Unique direction - k'], scenario.settings['Unique direction - l']]
            aoverb = scenario.settings['Ellipsoid a/b']
            if shape == 'Ellipsoid':
                direction = cell.convert_abc_to_xyz(hkl)
                depolarisation = Calculator.initialise_ellipsoid_depolarisation_matrix(direction,aoverb)
            elif shape == 'Plate':
                direction = cell.convert_hkl_to_xyz(hkl)
                depolarisation = Calculator.initialise_plate_depolarisation_matrix(direction)
            elif shape == 'Needle':
                direction = cell.convert_abc_to_xyz(hkl)
                depolarisation = Calculator.initialise_needle_depolarisation_matrix(direction)
            else:
                depolarisation = Calculator.initialise_sphere_depolarisation_matrix()
                direction = np.array( [] )
            direction = direction / np.linalg.norm(direction)
            self.directions.append(direction)
            #debugger.print('direction',direction)
            self.depolarisations.append(depolarisation)
        # Set up the parallel processing requirements before looping over the frequencies
        call_parameters = []
        for v in np.arange(float(vmin), float(vmax)+0.5*float(vinc), float(vinc)):
            vau = v * wavenumber
            call_parameters.append( (v, vau, mode_list, frequencies, sigmas, oscillator_strengths,
                                     volume, epsilon_inf, drude, drude_plasma, drude_sigma) )

        maximum_progress = len(call_parameters) * (1 + len(self.scenarios))
        progress = 0
        self.progressbar.setMaximum(maximum_progress)
        if self.notebook.progressbar is not None:
            self.notebook.progressbar.setMaximum(maximum_progress)
        QCoreApplication.processEvents()
        # Lets find ouit how many processors to use
        if self.notebook.ncpus == 0:
            number_of_processors = cpu_count()
        else:
            number_of_processors = self.notebook.ncpus
        #jk start = time.time()
        if self.notebook.threading:
            from multiprocessing.dummy import Pool
            pool = Pool(number_of_processors)
        else:
            from multiprocessing import Pool
            pool = Pool(number_of_processors, initializer=set_affinity_on_worker, maxtasksperchild=10)
        dielecv_results = []
        for result in pool.imap(Calculator.parallel_dielectric, call_parameters,chunksize=40):
            dielecv_results.append(result)
            progress += 1
            self.progressbar.setValue(progress)
            if self.notebook.progressbar is not None:
                self.notebook.progressbar.setValue(progress)
        pool.close()
        pool.join()
        QCoreApplication.processEvents()
        #jk print('Dielec calculation duration ', time.time()-start)
        nplots = len(dielecv_results)
        # Allocate space for the shared memory, we need twice as much as we have a complex data type
        shared_array_base = Array(ctypes.c_double, 2*nplots*9)
        previous_solution_shared = np.ctypeslib.as_array(shared_array_base.get_obj())
        # Convert the space allocated to complex
        previous_solution_shared.dtype = np.complex128
        # Reshape the array and fill everything with zero's
        previous_solution_shared = previous_solution_shared.reshape(nplots, 3,3)
        previous_solution_shared.fill(0.0+0.0j)
        # Prepare parallel call parameters for the loop over frequencies, methods, volume fractions
        concentration = self.settings['concentration']
        # Assemble each scenario settings
        self.xaxes = []
        self.realPermittivities = []
        self.imagPermittivities = []
        self.absorptionCoefficients = []
        self.molarAbsorptionCoefficients = []
        self.sp_atrs = []
        #jk start = time.time()
        if self.notebook.threading:
            from multiprocessing.dummy import Pool
            pool = Pool(number_of_processors)
        else:
            from multiprocessing import Pool
            pool = Pool(number_of_processors, initializer=set_affinity_on_worker, maxtasksperchild=10)
        for i,(scenario,L) in enumerate(zip(self.scenarios,self.depolarisations)):
            #debugger.print('Scenario ',i,L)
            matrix = scenario.settings['Matrix']
            method = scenario.settings['Effective medium method'].lower()
            matrix_density = scenario.settings['Matrix density']
            matrix_permittivity = np.identity(3) * scenario.settings['Matrix permittivity']
            mass_fraction = scenario.settings['Mass fraction']
            volume_fraction = scenario.settings['Volume fraction']
            particle_size_mu = scenario.settings['Particle size(mu)']
            particle_sigma_mu = scenario.settings['Particle size distribution sigma(mu)']
            shape = scenario.settings['Particle shape'].lower()
            hkl = [scenario.settings['Unique direction - h'], scenario.settings['Unique direction - k'], scenario.settings['Unique direction - l']]
            atr_refractive_index = scenario.settings['ATR material refractive index']
            atr_theta = scenario.settings['ATR theta']
            atr_spolfraction = scenario.settings['ATR S polarisation fraction']
            aoverb = scenario.settings['Ellipsoid a/b']
            bubble_vf = scenario.settings['Bubble volume fraction']
            bubble_radius = scenario.settings['Bubble radius']
            vf_type = ''
            call_parameters = []
            nplot = 0
            for v,vau,dielecv in dielecv_results:
                # convert the size to a dimensionless number which is 2*pi*size/lambda
                lambda_mu = 1.0E4 / (v + 1.0e-12)
                if particle_size_mu < 1.0e-12:
                    particle_size_mu = 1.0e-12
                size = 2.0*PI*particle_size_mu / lambda_mu
                data = ''
                call_parameters.append( (v,vau,dielecv,method,volume_fraction,vf_type,particle_size_mu,particle_sigma_mu,size,nplot,
                                         matrix_permittivity,shape,data,L,concentration,atr_refractive_index,atr_theta,atr_spolfraction,bubble_vf,bubble_radius,previous_solution_shared) )
                nplot += 1
            results = []
            for result in pool.imap(Calculator.solve_effective_medium_equations, call_parameters, chunksize=40):
                results.append(result)
                progress += 1
                self.progressbar.setValue(progress)
                if self.notebook.progressbar is not None:
                    self.notebook.progressbar.setValue(progress)
            QCoreApplication.processEvents()
            xaxis = []
            realPermittivity = []
            imagPermittivity = []
            absorptionCoefficient = []
            molarAbsorptionCoefficient = []
            sp_atr = []
            for v,nplot,method,vf_type,size_mu,size_sigma,shape,data,trace, absorption_coefficient,molar_absorption_coefficient,spatr in results:
                 xaxis.append(v)
                 realPermittivity.append(np.real(trace))
                 imagPermittivity.append(np.imag(trace))
                 absorptionCoefficient.append(absorption_coefficient)
                 molarAbsorptionCoefficient.append(molar_absorption_coefficient)
                 sp_atr.append(spatr)
            self.xaxes.append(xaxis)
            self.realPermittivities.append(realPermittivity)
            self.imagPermittivities.append(imagPermittivity)
            self.absorptionCoefficients.append(absorptionCoefficient)
            self.molarAbsorptionCoefficients.append(molarAbsorptionCoefficient)
            self.sp_atrs.append(sp_atr)
        pool.close()
        pool.join()
        #jk print('Dielec calculation duration ', time.time()-start)
        #if self.notebook.spreadsheet is not None:
        #    self.write_spreadsheet()
        self.dirty = False
        QApplication.restoreOverrideCursor()
        QCoreApplication.processEvents()

    def write_spreadsheet(self):
        if self.notebook.spreadsheet is None:
            return
        # Deal with Scenarios first
        sp = self.notebook.spreadsheet
        sp.selectWorkSheet('Scenarios')
        sp.delete()
        sp.writeNextRow(['A list of the scenarios used the calculation of the effective medium'],col=1)
        for index,(scenario,direction,depolarisation) in enumerate(zip(self.scenarios,self.directions,self.depolarisations)):
            sp.writeNextRow([''],col=1)
            sp.writeNextRow(['Scenario '+str(index)],col=1,check=1)
            settings = scenario.settings
            for key in sorted(settings,key=str.lower):
                sp.writeNextRow([key, settings[key]],col=1,check=1)
            sp.writeNextRow(['Normalised unique direction']+direction.tolist(), col=1,check=1)
            sp.writeNextRow(['Depolarisation matrix'], col=1,check=1)
            sp.writeNextRow(depolarisation[0].tolist(), col=2, check=1)
            sp.writeNextRow(depolarisation[1].tolist(), col=2, check=1)
            sp.writeNextRow(depolarisation[2].tolist(), col=2, check=1)
        # Now deal with Molar absorption, absorption, real and imaginary permittivity
        self.write_results(sp, 'Molar Absorption',self.xaxes, self.molarAbsorptionCoefficients)
        self.write_results(sp, 'Absorption',self.xaxes, self.absorptionCoefficients)
        self.write_results(sp, 'Real Permittivity',self.xaxes, self.realPermittivities)
        self.write_results(sp, 'Imaginary Permittivity',self.xaxes, self.imagPermittivities)
        self.write_results(sp, 'ATR Reflectance',self.xaxes, self.sp_atrs)

    def write_results(self, sp, name, vss, yss):
        sp.selectWorkSheet(name)
        sp.delete()
        headers = ['frequencies (cm-1)']
        #for isc,ys in enumerate(yss):
        #    headers.append('Scenario'+str(isc))
        headers.extend(self.legends)
        sp.writeNextRow(headers,row=0, col=1)
        for iv,v in enumerate(vss[0]):
           output = [v]
           for ys in yss:
               output.append(ys[iv])
           sp.writeNextRow(output, col=1,check=1)

    def plot(self,xs,ys,ylabel):
        # import matplotlib.pyplot as pl
        # mp.use('Qt5Agg')
        self.subplot = None
        self.remember_xs = xs
        self.remember_ys = ys
        self.remember_ylabel = ylabel
        self.figure.clf()
        if self.frequency_units == 'wavenumber':
            xlabel = r'Frequency $\mathdefault{(cm^{-1})}}$'
            scale = 1.0
        else:
            xlabel = r'THz'
            scale = 0.02998
        self.subplot = self.figure.add_subplot(111)
        for scenario,x,y in zip(self.scenarios,xs,ys):
            x = np.array(x)
            legend = scenario.settings['Legend']
            line, = self.subplot.plot(scale*x,y,lw=2, label=legend )
        self.subplot.set_xlabel(xlabel)
        self.subplot.set_ylabel(ylabel)
        self.subplot.legend(loc='best')
        self.subplot.set_title(self.settings['Plot title'])
        self.canvas.draw_idle()

    def replot(self):
        if self.subplot is not None:
            self.plot(self.remember_xs, self.remember_ys, self.remember_ylabel)





import sys
import os.path
from PyQt5.QtWidgets  import  QWidget
from PyQt5.QtWidgets  import  QListWidget, QComboBox, QLabel, QLineEdit
from PyQt5.QtWidgets  import  QFileDialog, QPushButton, QCheckBox
from PyQt5.QtWidgets  import  QFormLayout, QApplication
from PyQt5.QtWidgets  import  QVBoxLayout, QMessageBox
from PyQt5.QtCore     import  Qt, QCoreApplication
from PDielec.Utilities import  get_reader, Debug
from PDielec.GUI.SpreadSheetManager import SpreadSheetManager
import numpy as np

class MainTab(QWidget):

    def __init__(self, parent, program, filename, excelfile, debug=False):
        super(QWidget, self).__init__(parent)
        global debugger
        debugger = Debug(debug, 'MainTab')
        debugger.print('Initialising ',program, filename,excelfile)
        self.debug = debug
        self.settings = {}
        if program != '':
            self.settings['Program'] = program.lower()
        else:
            self.settings['Program'] = 'castep'
        self.settings['Output file name'] = filename
        self.directory = os.path.dirname(self.settings['Output file name'])
        self.settings['Excel file name'] = excelfile
        self.settings['QM program'] = ''
        self.settings['Hessian symmetrisation'] = 'symm'
        self.notebook = parent
        self.notebook.plottingCalculationRequired = True
        self.notebook.analysisCalculationRequired = True
        self.notebook.visualerCalculationRequired = True
        self.reader = None
        self.frequencies_cm1 = None
        self.dirty = True
        # Create first tab - MAIN
        vbox = QVBoxLayout()
        form = QFormLayout()
        #
        # The program combobox
        #
        self.program_cb = QComboBox(self)
        self.program_cb.setToolTip('Choose QM/MM program')
        self.program_cb.addItem('Abinit')
        self.program_cb.addItem('Castep')
        self.program_cb.addItem('Crystal')
        self.program_cb.addItem('Vasp')
        self.program_cb.addItem('Phonopy - VASP')
        self.program_cb.addItem('Quantum Espresso')
        self.program_cb.addItem('Gulp')
#        self.program_cb.addItem('Phonopy - QE')
#        self.program_cb.addItem('Phonopy - Crystal')
#        self.program_cb.addItem('Experiment')
        prtext = self.settings['Program'].capitalize()
        qmtext = self.settings['QM program'].capitalize()
        if prtext == 'Qe':
            prtext = 'Quantum Espresso'
        elif prtext == 'Vasp':
            prtext = 'VASP'
            self.settings['QM program'] = 'vasp'
        elif prtext == 'Phonopy':
            if qmtext != '':
                prtext = prtext + ' - ' + qmtext
        index = self.program_cb.findText(prtext, Qt.MatchFixedString)
        if index >=0:
            self.program_cb.setCurrentIndex(index)
        elif self.settings['Program'] == 'phonopy':
            index = 4
            self.settings['QM program'] = 'vasp'
            self.program_cb.setCurrentIndex(index)
        else:
            self.settings['Program'] = 'castep'
            index = 1
            self.program_cb.setCurrentIndex(index)
        self.program_cb.activated.connect(self.on_program_cb_activated)
        label = QLabel('QM/MM Program ')
        label.setToolTip('Choose QM/MM program')
        form.addRow(label, self.program_cb)
        #
        # The file selector
        #
        self.file_le = QLineEdit(self)
        self.file_le.setToolTip('Choose output file for analysis (press return for a file chooser)')
        self.file_le.setText(self.settings['Output file name'])
        self.file_le.returnPressed.connect(self.on_file_le_return)
        self.file_le.textChanged.connect(self.on_file_le_changed)
        label = QLabel('Analyse this output file')
        label.setToolTip('Choose output file for analysis (press return for a file chooser)')
        form.addRow(label, self.file_le)
        #
        # The method for symmetrising the hessian
        #
        self.hessian_symmetry_cb = QCheckBox(self)
        self.hessian_symmetry_cb.setToolTip('The Crystal program uses a different method for symmetrising the hessian.  Check this flag if you want to use the same method as used by Crystal14')
        self.hessian_symmetry_cb.setText('')
        self.hessian_symmetry_cb.setLayoutDirection(Qt.RightToLeft)
        if self.settings['Program'] == 'crystal':
            self.settings['Hessian symmetrisation'] = 'crystal'
            self.hessian_symmetry_cb.setCheckState(Qt.Checked)
            self.hessian_symmetry_cb.setEnabled(True)
        else:
            self.hessian_symmetry_cb.setCheckState(Qt.Unchecked)
            self.hessian_symmetry_cb.setEnabled(False)
        self.hessian_symmetry_cb.stateChanged.connect(self.on_hessian_symmetry_changed)
        label = QLabel('Symmetrise the hessian for Crystal14')
        label.setToolTip('The Crystal program uses a different method for symmetrising the hessian.  Check this flag if you want to use the same method as used by Crystal14')
        form.addRow(label, self.hessian_symmetry_cb)
        #
        # The store results
        #
        self.resultsfile_le = QLineEdit(self)
        self.resultsfile_le.setToolTip('Provide the name of an .xlsx file if results are to be stored in a spreadsheet')
        self.resultsfile_le.setText(self.settings['Excel file name'])
        self.resultsfile_le.returnPressed.connect(self.on_resultsfile_le_return)
        self.resultsfile_le.textChanged.connect(self.on_resultsfile_le_changed)
        label = QLabel('Excel spread sheet')
        label.setToolTip('Provide the name of an .xlsx file if results are to be stored in a spreadsheet')
        form.addRow(label, self.resultsfile_le)
        #
        # Calculation button
        #
        self.calculation_button = QPushButton('Read the output file and start the calculation')
        self.calculation_button.setToolTip('Pressing this button initiates reading of the output file and processing of all the options')
        self.calculation_button.clicked.connect(self.on_calculation_button_clicked)
        #label = QLabel('Read and calculate')
        #label.setToolTip('Pressing this button initiates reading of the output file and processing of all the options')
        form.addRow(self.calculation_button)
        # add form layout
        vbox.addLayout(form)
        # output window
        self.listw_l = QLabel('Frequencies from '+self.settings['Output file name'], self)
        vbox.addWidget(self.listw_l)
        self.listw = QListWidget(self)
        vbox.addWidget(self.listw)
        # finalise the layout
        self.setLayout(vbox)
        QCoreApplication.processEvents()
        # If the filename was given then force it to be read and processed
        if filename != '':
            debugger.print('Reading output file in maintab initialisation')
            self.on_calculation_button_clicked()
        QCoreApplication.processEvents()

    def on_calculation_button_clicked(self):
        debugger.print('Calculation button clicked')
        if len(self.settings['Excel file name']) > 5 and self.settings['Excel file name'][-5:] == '.xlsx':
            self.directory = os.path.dirname(self.settings['Output file name'])
            # open the file name with the directory of the output file name
            self.openSpreadSheet(os.path.join(self.directory,self.settings['Excel file name']))
        elif len(self.settings['Excel file name']) > 1 and self.settings['Excel file name'][-5:] != '.xlsx':
            # The file isn't valid so tell the user there is a problem
            QMessageBox.about(self,'Spreadsheet name','File name of spreadsheet must end in  .xlsx')
            return
        #
        # Read the output file
        #
        self.read_output_file()
        self.dirty = False
        # If there is a spread sheet write the spread sheet
        # if self.notebook.spreadsheet is not None:
        #     self.write_spreadsheet()

    def write_spreadsheet(self):
        sp = self.notebook.spreadsheet
        if sp is None:
            return
        debugger.print('Reading output file ', self.settings['Output file name'])
        sp.selectWorkSheet('Main')
        sp.delete()
        debugger.print('write_spreadsheet',self.settings)
        sp.writeNextRow( ['Main Tab Settings'], col=1 )
        for item in sorted(self.settings):
            sp.writeNextRow([item,self.settings[item]], col=1, check=1)
        sp.writeNextRow( [''], col=1 )
        sp.writeNextRow( ['Frequencies (cm1) as read from the output file'], col=1 )
        for ifreq in enumerate(self.frequencies_cm1):
            sp.writeNextRow(ifreq, col=1, check=1)

    def read_output_file(self):
        if self.settings['Output file name'] == '':
            return
        if not os.path.isfile(self.settings['Output file name']):
            QMessageBox.about(self,'Processing output file','The filename for the output file to be processed is not correct: '+self.settings['Output file name'])
            return
        debugger.print('Read output file - clear list widget')
        self.listw.clear()
        self.listw_l.setText('Frequencies from '+self.settings['Output file name'])
        self.reader = get_reader(self.settings['Program'],[ self.settings['Output file name'] ], self.settings['QM program'] )
        if self.reader is None:
            print('Error in reading files - program  is ',self.settings['Program'])
            print('Error in reading files - filename is ',self.settings['Output file name'])
            print('Need to choose the file and program properly')
            QMessageBox.about(self,'Processing output file','A reader has not be created for this filename: '+self.settings['Output file name'])
            return
        #switch on debugging in the reader
        #self.reader.debug = self.debug
        self.reader.hessian_symmetrisation = self.settings['Hessian symmetrisation']
        try:
            self.reader.read_output()
        except:
            print('Error in reading output files - program  is ',self.settings['Program'])
            print('Error in reading output files - filename is ',self.settings['Output file name'])
            print('Need to choose the file and program properly')
            QMessageBox.about(self,'Processing output file','Error on reading the output file using read_output(): '+self.settings['Output file name'])
            return
        if len(self.reader.unit_cells) == 0:
            print('Error in reading output files - program  is ',self.settings['Program'])
            print('Error in reading output files - filename is ',self.settings['Output file name'])
            print('Need to choose the file and program properly')
            QMessageBox.about(self,'Processing output file','The output file has no unit cells in it: '+self.settings['Output file name'])
            return
        QApplication.setOverrideCursor(Qt.WaitCursor)
        # Update the checkbox
        if self.settings['Hessian symmetrisation'] == 'crystal':
            self.hessian_symmetry_cb.setCheckState(Qt.Checked)
        else:
            self.hessian_symmetry_cb.setCheckState(Qt.Unchecked)
        if self.settings['Program'] == 'crystal':
            self.hessian_symmetry_cb.setEnabled(True)
        else:
            self.hessian_symmetry_cb.setEnabled(False)
        # tell the notebook that we have read the info and we have a reader
        self.notebook.reader = self.reader
        self.notebook.plottingCalculationRequired = True
        self.notebook.analysisCalculationRequired = True
        self.notebook.visualerCalculationRequired = True
        self.directory = os.path.dirname(self.settings['Output file name'])
        if self.debug:
            self.reader.print_info()
        self.frequencies_cm1 = np.sort(self.reader.frequencies)
        for f in self.frequencies_cm1:
            self.listw.addItem('{0:.3f}'.format(f))
        # tell the settings tab to update the widgets that depend on the contents of the reader
        debugger.print('processing a return')
        if hasattr(self.notebook, 'settingsTab'):
            debugger.print('about to refresh settings')
            self.notebook.settingsTab.refresh(True)
        # Update any scenarios
        if hasattr(self.notebook, 'scenarios'):
            debugger.print('about to refresh scenarios')
            debugger.print('notebook has {} scenarios'.format(len(self.notebook.scenarios)))
            for tab in self.notebook.scenarios:
                tab.refresh()
        else:
            debugger.print('notebook has no scenarios yet')
        # Update the plotting tab
        if hasattr(self.notebook, 'plottingTab'):
            debugger.print('about to refresh plottingtab')
            self.notebook.plottingTab.refresh()
        else:
            debugger.print('notebook has no plotting tab yet')
        # Update the analysis tab
        if hasattr(self.notebook, 'analysisTab'):
            debugger.print('about to refresh analysisTab')
            self.notebook.analysisTab.refresh()
        else:
            debugger.print('notebook has no analysis tab yet')
        # Update the viewer tab
        if hasattr(self.notebook, 'viewerTab'):
            debugger.print('about to refresh viewerTab')
            self.notebook.viewerTab.refresh()
        else:
            debugger.print('notebook has no viewer tab yet')
        # Update the fitter tab
        if hasattr(self.notebook, 'fitterTab'):
            debugger.print('about to refresh fitterTab')
            self.notebook.fitterTab.refresh()
        else:
            debugger.print('notebook has no fitter tab yet')
        QApplication.restoreOverrideCursor()

    def on_hessian_symmetry_changed(self):
        debugger.print('on hessian symmetry changed')
        if self.hessian_symmetry_cb.isChecked():
            self.settings['Hessian symmetrisation'] = 'crystal'
        else:
            self.settings['Hessian symmetrisation'] = 'symm'
        self.dirty = True

    def on_resultsfile_le_return(self):
        debugger.print('on resultsfile return ', self.resultsfile_le.text())

    def on_resultsfile_le_changed(self, text):
        debugger.print('on resultsfile changed', text)
        text = self.resultsfile_le.text()
        self.settings['Excel file name'] = text
        self.dirty = True

    def openSpreadSheet(self,text):
        if self.notebook.spreadsheet is not None:
            self.notebook.spreadsheet.close()
        if text[-5:] == '.xlsx':
            self.notebook.spreadsheet = SpreadSheetManager(text)
            self.dirty = True
        else:
           print('spreadsheet name not valid', text)
           exit()

    def on_file_le_return(self):
        debugger.print('on file return ', self.file_le.text())
        # Does the file exist?
        self.settings['Output file name'] = self.file_le.text()
        # Open a file chooser
        #options = QFileDialog.Options()
        #options |= QFileDialog.DontUseNativeDialog
        filename, _ = QFileDialog.getOpenFileName(self,'Open MM/QM Output file','','Castep (*.castep);;Abinit (*.out);;Gulp (*.gout);;VASP (OUTCAR*);; QE (*.dynG);; Crystal 14 (*.out);; Phonopy (OUTCAR*);; All Files (*)')
        if filename != '':
            self.settings['Output file name'] = filename
            self.file_le.setText(self.settings['Output file name'])
            debugger.print('new file name', self.settings['Output file name'])
            self.directory = os.path.dirname(self.settings['Output file name'])
            self.dirty = True
        return

    def on_file_le_changed(self, text):
        debugger.print('on file changed', text)
        self.settings['Output file name'] = text
        self.dirty = True

    def on_program_cb_activated(self,index):
        debugger.print('on program combobox activated', index)
        debugger.print('on program combobox activated', self.program_cb.currentText())
        text = self.program_cb.currentText()
        self.settings['Program'] = text.lower()
        self.settings['QM program'] = ''
        if text == 'Phonopy - VASP':
            self.settings['Program']   = 'phonopy'
            self.settings['QM program'] = 'vasp'
        elif text == 'Phonopy - QE':
            self.settings['Program']   = 'phonopy'
            self.settings['QM program'] = 'qe'
        elif text == 'Phonopy - Crystal':
            self.settings['Program']   = 'phonopy'
            self.settings['QM program'] = 'crystal'
        elif text == 'Quantum Espresso':
            self.settings['Program']   = 'qe'
        debugger.print('Program is now  ', self.settings['Program'])
        debugger.print('QM program is now', self.settings['QM program'])
        if self.settings['Program'] == 'crystal':
            self.hessian_symmetry_cb.setEnabled(True)
            self.hessian_symmetry_cb.setCheckState(Qt.Checked)
        else:
            self.hessian_symmetry_cb.setEnabled(True)
            self.hessian_symmetry_cb.setCheckState(Qt.Unchecked)
            self.hessian_symmetry_cb.setEnabled(False)
        self.dirty = True

    def refresh(self,force=False):
        debugger.print('refresh', force)
        if not self.dirty and not force:
            return
        self.notebook.newPlottingCalculationRequired = True
        self.notebook.newAnalysisCalculationRequired = True
        self.notebook.newVisualerCalculationRequired = True
        #
        # Block signals during refresh
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(True)
        prtext = self.settings['Program'].capitalize()
        qmtext = self.settings['QM program'].capitalize()
        if prtext == 'Qe':
            prtext = 'Quantum Espresso'
        elif prtext == 'Vasp':
            prtext = 'VASP'
            self.settings['QM program'] = 'vasp'
        elif prtext == 'Phonopy':
            if qmtext != '':
                prtext = prtext + ' - ' + qmtext
        index = self.program_cb.findText(prtext, Qt.MatchFixedString)
        if index >=0:
            self.program_cb.setCurrentIndex(index)
        elif self.settings['Program'] == 'phonopy':
            index = 4
            self.settings['QM program'] = 'vasp'
            self.program_cb.setCurrentIndex(index)
        else:
            self.settings['Program'] = 'castep'
            index = 1
            self.program_cb.setCurrentIndex(index)
        if self.settings['Hessian symmetrisation'] == 'crystal':
            self.hessian_symmetry_cb.setCheckState(Qt.Checked)
        else:
            self.hessian_symmetry_cb.setCheckState(Qt.Unchecked)
        if self.settings['Program'] == 'crystal':
            self.hessian_symmetry_cb.setEnabled(True)
        else:
            self.hessian_symmetry_cb.setEnabled(False)
        self.file_le.setText(self.settings['Output file name'])
        self.resultsfile_le.setText(self.settings['Excel file name'])
        self.on_calculation_button_clicked()
        self.dirty = False
        #
        # UnBlock signals
        #
        for w in self.findChildren(QWidget):
            w.blockSignals(False)
        return



#!/usr/bin/python
#
# Copyright 2015 John Kendrick
#
# This file is part of PDielec
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# You should have received a copy of the MIT License
# along with this program, if not see https://opensource.org/licenses/MIT
#
"""Read the contents of a directory containing Abinit input and output files
   Inherit the following from the GenericOutputReader
   __init__
   printInfo
   _read_output_file
"""

import re
import math
import numpy as np
from PDielec.Constants import amu, angs2bohr, atomic_number_to_element, hartree2ev
from PDielec.UnitCell import UnitCell
from PDielec.GenericOutputReader import GenericOutputReader


class AbinitOutputReader(GenericOutputReader):
    """Read the contents of a directory containing Abinit input and output files
       Inherit the following from the GenericOutputReader
       __init__
       printInfo
       _read_output_file
"""

    def __init__(self, filenames):
        GenericOutputReader.__init__(self, filenames)
        self.type = 'Abinit output files'
        self._acell = None
        self._charges = None
        return

    def _read_output_files(self):
        """Read the Abinit file names"""
        # Define the search keys to be looked for in the files
        self.manage = {}   # Empty the dictionary matching phrases
        self.manage['dynamical']  = (re.compile('  Dynamical matrix,'), self._read_dynamical)
        self.manage['bornCharges']  = (re.compile('  Effective charges,'), self._read_born_charges)
        self.manage['epsilon']  = (re.compile('  Dielectric tensor,'), self._read_epsilon)
        self.manage['masses']   = (re.compile('              amu '), self._read_masses)
        self.manage['nions']    = (re.compile('            natom '), self._read_natom)
        self.manage['lattice']  = (re.compile('            rprim '), self._read_lattice_vectors)
        self.manage['xred']  = (re.compile('             xred '), self._read_xred)
        self.manage['typat']    = (re.compile('            typat '), self._read_typat)
        self.manage['ntypat']   = (re.compile('           ntypat '), self._read_ntypat)
        self.manage['acell']    = (re.compile('            acell '), self._read_acell)
        self.manage['nkpt']    = (re.compile('             nkpt '), self._read_kpoints)
        self.manage['band']    = (re.compile('            nband '), self._read_band)
        self.manage['band1']    = (re.compile('            nband1 '), self._read_band)
        self.manage['occupancy']    = (re.compile('              occ '), self._read_occupancy)
        self.manage['occupancy1']    = (re.compile('              occ1 '), self._read_occupancy)
        self.manage['ecut']    = (re.compile('^ *ecut '), self._read_energy_cutoff)
        self.manage['kptrlatt']    = (re.compile('         kptrlatt '), self._read_kpoint_grid)
        self.manage['electrons']    = (re.compile('  fully or partial'), self._read_electrons)
        self.manage['pressure']    = (re.compile('-Cartesian.*GPa'), self._read_pressure)
        self.manage['znucl']    = (re.compile('^  *znucl '), self._read_znucl)
        self.manage['totalenergy']    = (re.compile('^  *Total energy '), self._read_total_energy)
        for f in self._outputfiles:
            self._read_output_file(f)
        return

    def _read_total_energy(self, line):
        self.final_energy_without_entropy = float(line.split()[4]) * hartree2ev
        self.final_free_energy = float(line.split()[4]) * hartree2ev
        return

    def _read_znucl(self, line):
        self.species = []
        for z in line.split()[1:]:
            iz = int(float(z)+0.001)
            self.species.append(atomic_number_to_element[iz].capitalize())
        self.nspecies = len(self.species)
        return

    def _read_band(self, line):
        self.nbands = int(line.split()[1])
        return

    def _read_occupancy(self, line):
        occs = []
        occupancies = line.split()[1:]
        while len(occs) < self.nbands:
            occs+= [ float(f) for f in occupancies ]
            occupancies = self.file_descriptor.readline().split()
        sum = 0.0
        for f in occs:
            sum += f
        self.electrons = int(sum + 0.0001)
        return

    def _read_pressure(self, line):
        self.pressure = float(line.split()[7])
        return

    def _read_electrons(self, line):
        self.electrons = float(line.split()[6])
        return

    def _read_kpoint_grid(self, line):
        self.kpoint_grid = [ int(line.split()[1]), int(line.split()[5]), int(line.split()[9]) ]
        return

    def _read_kpoints(self, line):
        self.kpoints = int(line.split()[1])
        return

    def _read_energy_cutoff(self, line):
        self.energy_cutoff = hartree2ev * float(line.split()[1])
        return

    def _read_acell(self, line):
        self._acell = [float(f)/angs2bohr for f in line.split()[1:4]]
        return

    def _read_ntypat(self, line):
        self.nspecies = int(line.split()[1])
        return

    def _read_typat(self, line):
        # typat occurs last in the list of data items we need from the output file
        self.atom_type_list = [int(i)-1 for i in line.split()[1:]]
        self.masses = [None for i in range(self.nions)]
        self.ions_per_type = [0 for i in range(self.nspecies)]
        for i, a in enumerate(self.atom_type_list):
            self.ions_per_type[a-1] += 1
            self.masses[i] = self.masses_per_type[a]
        if self.species:
            species_list = [ self.species[i] for i in self.atom_type_list ]
            self.unit_cells[-1].set_element_names(species_list)
        return

    def _read_epsilon(self, line):
        for i in range(3):
            linea = self.file_descriptor.readline().split()
        nlines = 9
        for i in range(nlines):
            linea = self.file_descriptor.readline().split()
            if not linea:
                linea = self.file_descriptor.readline().split()
            j = int(linea[0])
            k = int(linea[2])
            self.zerof_optical_dielectric[j-1][k-1] = float(linea[4])
        return

    def _read_natom(self, line):
        self.nions = int(line.split()[1])
        # We can only create this once we know the number of ions
        self._charges = np.zeros((self.nions, 3, 3))
        return

    def _read_masses(self, line):
        self.masses_per_type = [float(f) for f in line.split()[1:]]
        return

    def _read_dynamical(self, line):
        # Read the dynamical matrix
        nmodes = self.nions*3
        hessian = np.zeros((nmodes, nmodes))
        for i in range(4):
            self.file_descriptor.readline()
        nlines = nmodes*nmodes
        for i in range(nlines):
            linea = self.file_descriptor.readline().split()
            if not linea:
                linea = self.file_descriptor.readline().split()
            diri  = int(linea[0])
            atomi = int(linea[1])
            dirj  = int(linea[2])
            atomj = int(linea[3])
            ipos  = (atomi - 1)*3 + diri - 1
            jpos  = (atomj - 1)*3 + dirj - 1
            # store the massweighted matrix
            hessian[ipos][jpos] = float(linea[4])/(amu*math.sqrt(self.masses[atomi-1]*self.masses[atomj-1]))
        # symmetrise, project diagonalise and store frequencies and normal modes
        self._dynamical_matrix(hessian)
        return

    def _read_born_charges(self, line):
        """Read the born charges from the outputfile file.
           Each row of the output refers to a given field direction
           Each column in the row refers the atomic displacement
           so the output is arranged [[a1x a1y a1z]
                                      [a2x a2y a2z]
                                      [a3x a3y a3z]]
           where 1,2,3 are the field directions and x, y, z are the atomic displacements"""
        for i in range(5):
            self.file_descriptor.readline()
        #  The charges are calculated in two ways, we take the mean of the phonon and the field
        nlines = 9*self.nions
        for i in range(nlines):
            linea = self.file_descriptor.readline().split()
            if not linea:
                linea = self.file_descriptor.readline().split()
            if int(linea[3]) > self.nions:
                ifield = int(linea[2])
                ixyz   = int(linea[0])
                iatom  = int(linea[1])
            else:
                ifield = int(linea[0])
                ixyz   = int(linea[2])
                iatom  = int(linea[3])
            self._charges[iatom-1][ifield-1][ixyz-1] += 0.5*float(linea[4])
        # Convert the charges
        self.born_charges = []
        for i in range(self.nions):
            atom = []
            for ifield in range(3):
                b = self._charges[i][ifield][:].tolist()
                atom.append(b)
            self.born_charges.append(atom)
        return

    def _read_xred(self, line):
        linea = line.split()[1:]
        fractional = []
        fractional.append( [ float(xyz) for xyz in linea ] )
        for i in range(self.nions-1):
            linea = self.file_descriptor.readline().split()
            fractional.append( [ float(xyz) for xyz in linea ] )
        # end for i
        self.unit_cells[-1].set_fractional_coordinates(fractional)
        if self.species:
            species_list = [ self.species[i] for i in self.atom_type_list ]
            self.unit_cells[-1].set_element_names(species_list)
    # end def


    def _read_lattice_vectors(self, line):
        linea = line.split()
        avector = [float(linea[1]), float(linea[2]), float(linea[3])]
        linea = self.file_descriptor.readline().split()
        bvector = [float(linea[0]), float(linea[1]), float(linea[2])]
        linea = self.file_descriptor.readline().split()
        cvector = [float(linea[0]), float(linea[1]), float(linea[2])]
        avector = [f * self._acell[0] for f in avector]
        bvector = [f * self._acell[1] for f in bvector]
        cvector = [f * self._acell[2] for f in cvector]
        self.unit_cells.append(UnitCell(avector, bvector, cvector))
        self.ncells = len(self.unit_cells)
        self.volume = self.unit_cells[-1].volume
        return



#!/usr/bin/env python
"""Read in two csv files and check that the float contents are the same"""
from __future__ import print_function
import string
import os, sys
from os.path import splitext
from termcolor import colored
from openpyxl import load_workbook

global threshold

# Start processing the directories
if len(sys.argv) <= 1 :
    print('checkexcel file file2 [-thresh 1.0E-3] [-f]', file=sys.stderr)
    print('         Compare the two excel files for any significant changes', file=sys.stderr)
    print('         Numbers f1 and f2 in each file are compared', file=sys.stderr)
    print('         an error is flagged if 2*abs(f1-f2)/(f1+f2+2) > threshold', file=sys.stderr)
    print('         The default value for threshold is 1.0E-3 ', file=sys.stderr)
    print('         -f forces a full comparison of the sheets ', file=sys.stderr)
    print('            by default Settings is not included    ', file=sys.stderr)
    exit()

threshold = 1.0E-3
separator = None
tokens = sys.argv[1:]
ntokens = len(tokens)-1
itoken = -1
files = []
full = False
while itoken < ntokens:
    itoken += 1
    token = tokens[itoken]
    if token == '-thresh':
        itoken +=1
        threshold = float(tokens[itoken])
    if token == '-f':
        full = True
    else:
        files.append(tokens[itoken])
    # end if
# end while

file1 = files[0]
file2 = files[1]

#
# Open the work books
#
wb1 = load_workbook(filename=file1, read_only=True)
wb2 = load_workbook(filename=file2, read_only=True)
#
# Initialise variables
#
error = None
nerrors = 0
max_percentage_error = 0.0
#
# Loop over sheets
#
sheets = ['Main','Scenarios','Molar Absorption','Absorption','Real Permittivity','Imaginary Permittivity', 'ATR Reflectance', 'Analysis']
if full:
    sheets.append('Settings')
for sheet in sheets:
    ws1 = wb1[sheet]
    ws2 = wb2[sheet]
    max_rows1 = ws1.max_row
    max_rows2 = ws2.max_row
    max_columns1 = ws1.max_column
    max_columns2 = ws2.max_column
    if max_rows1 != max_rows2:
        print('Error - the number rows in the sheet are not the same',sheet,max_rows1,max_rows2)
        nerrors += 1
        continue
    if max_columns1 != max_columns2:
        print('Error - the number columns in the sheet are not the same',sheet,max_columns1,max_columns2)
        nerrors += 1
        continue
    #
    # Loop over rows
    #
    row_index = 0
    for row1,row2 in zip(ws1.rows, ws2.rows):
        row_index += 1
        col_index = 0
        #
        # Loop over cells
        #
        for cell1,cell2, in zip(row1,row2):
            col_index += 1
            value1 = cell1.value
            value2 = cell2.value
            if cell1 is None and cell2 is None:
                pass
            elif cell1 is None or cell2 is None:
                nerrors += 1
                percentage_error = 0.0
                error = (sheet, row_index, col_index, value1, value2, percentage_error)
            else:
                if value1 != value2:
                    if cell1.data_type == 'n' and cell2.data_type == 'n':
                        #
                        # Flag an error which is numeric
                        #
                        percentage_error = 100.0*abs(2.0*(value1 - value2)/(value1 + value2+2))
                        if percentage_error > 100.*threshold:
                            nerrors += 1
                            if percentage_error > max_percentage_error:
                                max_percentage_error = percentage_error
                                error = (sheet, row_index, col_index, value1, value2, percentage_error)
                            # if percentage error
                        else:
                            if percentage_error > max_percentage_error:
                                max_percentage_error = percentage_error
                        # if percentage error > threshold
                    else:
                        #
                        # This is a non-numeric error
                        #
                        nerrors += 1
                        percentage_error = 0.0
                        error = (sheet, row_index, col_index, value1, value2, percentage_error)
                    # if cell1.data_type
                # if value1 != value2
            # if cell1 is none
        # for cell1, cell2
    # for rowq, row
# for sheet
if error is not None:
    sheet,row,col,value1,value2,max_percentage_error = error
    print('  '+colored('ERRORS:','red')+'({}) LARGEST ON ROW,COL {},{} OF SHEET {}, {}({}) and {}({}) -- max %error={}'.format(nerrors, row,col,sheet,file1,value1,file2,value2,max_percentage_error))
elif nerrors > 0:
    print('  '+colored('ERRORS:','red')+'({}) Dimensions of spreadsheet were wrong                                    '.format(nerrors))
else:
    print('  '+colored('OK:','blue')+" {} = {} -- max %error={}" .format(file1,file2,max_percentage_error))
